----------------------------------------------------------------------------
-------------------------      M a R T E   O S      ------------------------
----------------------------------------------------------------------------
--                                                            {MARTE_VERSION}
--
--                       'r t _ e p _ p r o t o c o l . a d b'
--
--                                     Ada
--
--
--  File 'rt_ep_protocol.adb'                                        By Chema.
--                                                         Jose Maria Martinez
--                                                        <martinjm@unican.es>
--  Import declarations and translations from rt_comm.h
--
--
--  {MARTE_COPYRIGHT}
--
-----------------------------------------------------------------------------
--  {<MAR}
--  Chema:
--  Date format DD-MM-YY:
--
--
--  15-09-03: First FINAL draft (First attempt was in 2002).
--
--
--  {MAR>} -------------------------------------------------------------------

with System;
--  with Ada.Streams; use Ada.Streams;
with Interfaces; use Interfaces;
--  with Ada.Unchecked_Conversion;
--  with Ada.Text_IO; use Ada.Text_IO;
package body Rt_Ep_Protocol  is


   --     type Data_Array_Type is array (Positive range 1..Max_Info_Size)
   --       of Interfaces.Unsigned_8;

   --     type Data_Array_Type_Ac is access all Data_Array_Type;

   --   type Ethernet_Address_Ac is access all Ethernet_Address;

   -----------------------
   --  rt_comm.h imports --
   -----------------------

   --  Send_Info_C

   procedure Send_Info_C (Result : out Integer;
                          Destination_Address : in out Ethernet_Address;
                          Channel_ID : in Unsigned_16;
                          Data : in System.Address;  --  char *
                          Length_Data : in Unsigned_16;
                          Data_Priority : in Unsigned_8);
   pragma Import (C, Send_Info_C, "send_info");
   pragma Import_Valued_Procedure (Send_Info_C, "send_info");

   --  Recv_Info_C
   --  Must  be in parameter cause we want to have the
   --  value not the reference of the System.Address.
   --  If we had put in out the value would be the
   --  reference from a System.Address which doesn't
   --  make any sense.
   procedure Recv_Info_C (Result : out Integer;
                          Source_Address : in out Ethernet_Address;
                          Channel_ID : in Unsigned_16;
                          Data : in System.Address;
                          Data_Priority : out Unsigned_8);
   pragma Import (C, Recv_Info_C, "recv_info");
   pragma Import_Valued_Procedure (Recv_Info_C, "recv_info");


   --  Any_Info_C
   function Any_Info_C (Channel_ID : in Unsigned_16) return Integer;
   pragma Import (C, Any_Info_C, "any_info");


   --  Init_Comm
   function Init_Comm_C (Prio_Ceiling : in Integer) return Integer;
   pragma Import (C, Init_Comm_C, "init_comm");


   --------------------------
   --  ADA RT-EP Interface. --
   --------------------------

   ---------------
   --  Send_Info --
   ---------------

   --  Raise on Error:
   --  Station_Not_Valid
   --  Station_Not_Found
   --  Invalid_Channel
   --  Synch_Error
   --  Unexpected_Error


   -----------------------
   --  Generic Send Info --
   -----------------------

   procedure Generic_Send_Info (Destination_Address : in out Ethernet_Address;
                                Channel_ID : in RT_EP_Channel_Number;
                                Data : in Data_Type;
                                Size : in RT_EP_Packet_Size;
                                Data_Priority : in RT_EP_Priority) is

      Result : Integer;
   begin


      Send_Info_C (Result, Destination_Address,
                   Unsigned_16 (Channel_ID), Data'Address, Unsigned_16 (Size),
                   Unsigned_8 (Data_Priority));
      case Result is
         when -20 =>
            raise Synch_Error;
         when -55 =>
            raise Unexpected_Error;
         when -10 =>
            raise Station_Not_Valid;
         when -9 =>
            raise Station_Not_Found;
         when -4 =>
            raise Invalid_Channel;
         when others =>
            null;
      end case;

   end Generic_Send_Info;


   ---------------
   --  Recv_Info --
   ---------------

   --  Raise on Error:
   --  Invalid_Channel
   --  Synch_Error
   --  Unexpected_Error


   -----------------------
   --  Generic Recv_Info --
   -----------------------

   procedure Generic_Recv_Info (Source_Address : in out Ethernet_Address;
                                Channel_ID : in RT_EP_Channel_Number;
                                Data : out Data_Type;
                                Size : out RT_EP_Packet_Size;
                                Data_Priority : out RT_EP_Priority) is

      Result : Integer;
      Data_Recv_Ac : System.Address;
   begin
      Data_Recv_Ac := Data'Address;
      Recv_Info_C (Result, Source_Address, Unsigned_16 (Channel_ID),
                   Data_Recv_Ac, Unsigned_8 (Data_Priority));

      case Result is
         when -20 =>
            raise Synch_Error;
         when -55 =>
            raise Unexpected_Error;
         when -4 =>
            raise Invalid_Channel;
         when others =>
            --  We catch the receiving element size.
            Size := RT_EP_Packet_Size (Result);
            --  null;
      end case;

   end Generic_Recv_Info;


   --------------
   -- Any_Info --
   --------------

   --  Raise on Error:
   --  Invalid_Channel
   --  Synch_Error
   --  Unexpected_Error

   function Any_Info (Channel_ID : in RT_EP_Channel_Number) return Boolean is

      Result : Integer;
   begin

      Result := Any_Info_C (Unsigned_16 (Channel_ID));

      case Result is
         when -20 =>
            raise Synch_Error;
         when -55 =>
            raise Unexpected_Error;
         when -4 =>
            raise Invalid_Channel;
         when -5 =>
            return False;
         when 0 =>
            return True;
         when others =>
            --  If all is working properly you wont get here.
            raise Unexpected_Error;
      end case;

   end Any_Info;


   ---------------
   -- Init_Comm --
   ---------------

   --  Raise on Error:
   --  Creation_Error
   --  Unexpected_Error

   function Init_Comm (Prio_Ceiling : Integer) return Integer is

      Result : Integer;
   begin

      Result := Init_Comm_C (Prio_Ceiling);

      case Result is
         when -55 =>
            raise Unexpected_Error;
         when -30 =>
            raise Creation_Error;
         when others =>
            null;
      end case;

      return Result;

   end Init_Comm;

--     -------------------------------------------------------------------
--     --- RT-EP with Streams  -------------------------------------------
--     -------------------------------------------------------------------


--     procedure Put_Function_ID (Stream : in out Function_Stub_Type;
--                                Function_ID : in Function_ID_Type) is
--     begin
--        Stream.Function_ID :=  Stream.Function_ID;
--     end Put_Function_ID;

--     function Get_Function_ID (Stream : in Function_Stub_Type)
--                              return Function_ID_Type is
--     begin
--        return Stream.Function_ID;
--     end Get_Function_ID;


--     procedure Stream_Stub (Function_Stub : in out  Function_Stub_Type;
--                            Function_Ptr : out Function_Stub_Type_Ac) is
--     begin

--        Function_Ptr := Function_Stub'Unchecked_Access;

--     end Stream_Stub;

--     ---------------
--     -- Recv_Info --
--     ---------------

--     --  Raise on Error:
--     --  Invalid_Channel
--     --  Synch_Error
--     --  Unexpected_Error

--     procedure Recv_Info (Source_Address : in out Ethernet_Address;
--                          Channel_ID : in RT_EP_Channel_Number;
--                          Data : in out Function_Stub_Type;
--                          Size : out RT_EP_Packet_Size;
--                          Data_Priority : out RT_EP_Priority)is

--        Result : Integer;
--        Data_Recv_Ac : System.Address;
--     begin
--        Data_Recv_Ac := Data'Address;
--        Recv_Info_C (Result, Source_Address, Unsigned_16 (Channel_ID),
--                     Data_Recv_Ac, Unsigned_8 (Data_Priority));

--        case Result is
--           when -20 =>
--              raise Synch_Error;
--           when -55 =>
--              raise Unexpected_Error;
--           when -4 =>
--              raise Invalid_Channel;
--           when others =>
--              --  We catch the receiving element size.
--              Size := RT_EP_Packet_Size (Result);
--              --  null;
--        end case;

--     end Recv_Info;

--     ---------------
--     -- Send_Info --
--     ---------------

--     --  Raise on Error:
--     --  Station_Not_Valid
--     --  Station_Not_Found
--     --  Invalid_Channel
--     --  Synch_Error
--     --  Unexpected_Error

--     procedure Send_Info (Destination_Address : in out Ethernet_Address;
--                          Channel_ID : in RT_EP_Channel_Number;
--                          Data : in out Function_Stub_Type;
--                          Size : in RT_EP_Packet_Size;
--                          Data_Priority : in RT_EP_Priority) is

--        Result : Integer;
--     begin


--        Send_Info_C (Result, Destination_Address,
--                Unsigned_16 (Channel_ID), Data'Address, Unsigned_16 (Size),
--                     Unsigned_8 (Data_Priority));
--        case Result is
--           when -20 =>
--              raise Synch_Error;
--           when -55 =>
--              raise Unexpected_Error;
--           when -10 =>
--              raise Station_Not_Valid;
--           when -9 =>
--              raise Station_Not_Found;
--           when -4 =>
--              raise Invalid_Channel;
--           when others =>
--              null;
--        end case;

--     end Send_Info;

end Rt_Ep_Protocol;
