------------------------------------------------------------------------------
--                                                                        --
--                            GLADE COMPONENTS                              --
--                                                                          --
--          S Y S T E M . G A R L I C . P R O T O C O L S . T C P           --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                            $Revision: 1.6 $
--                                                                          --
--         Copyright (C) 1996-2000 Free Software Foundation, Inc.           --
--                                                                          --
-- GARLIC is free software;  you can redistribute it and/or modify it under --
-- terms of the  GNU General Public License  as published by the Free Soft- --
-- ware Foundation;  either version 2,  or (at your option)  any later ver- --
-- sion.  GARLIC is distributed  in the hope that  it will be  useful,  but --
-- WITHOUT ANY WARRANTY;  without even the implied warranty of MERCHANTABI- --
-- LITY or  FITNESS FOR A PARTICULAR PURPOSE.  See the  GNU General Public  --
-- License  for more details.  You should have received  a copy of the GNU  --
-- General Public License  distributed with GARLIC;  see file COPYING.  If  --
-- not, write to the Free Software Foundation, 59 Temple Place - Suite 330, --
-- Boston, MA 02111-1307, USA.                                              --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable  to  be --
-- covered  by the  GNU  General  Public  License.  This exception does not --
-- however invalidate  any other reasons why  the executable file  might be --
-- covered by the  GNU Public License.                                      --
--                                                                          --
--               GLADE  is maintained by ACT Europe.                        --
--               (email: glade-report@act-europe.fr)                        --
--                                                                          --
------------------------------------------------------------------------------
with Text_IO; use Text_IO;

with Global_Priorities.Mapping;  use Global_Priorities.Mapping;  --  jl

--  Para Comparacion canales U_16
with Local_Protocol;  --  jl LOCAL_PROTOCOL;

with System.Garlic.Debug;                 use System.Garlic.Debug;
with System.Garlic.Exceptions;            use System.Garlic.Exceptions;
with System.Garlic.Heart;                 use System.Garlic.Heart;
with System.Garlic.Options;
with System.Garlic.Partitions;            use System.Garlic.Partitions;
with System.Garlic.Protocols;             use System.Garlic.Protocols;
with System.Garlic.Soft_Links;
with System.Garlic.Streams;               use System.Garlic.Streams;
with System.Garlic.Types;                 use System.Garlic.Types;
with System.Garlic.Utils;                 use System.Garlic.Utils;

with System.Storage_Elements;             use System.Storage_Elements;

package body System.Garlic.Protocols.Local is

   Private_Debug_Key : constant Debug_Key :=
     Debug_Initialize ("S_GAPRTC", "(s-gaprtc): ");

   procedure D
     (Message : in String;
      Key     : in Debug_Key := Private_Debug_Key)
     renames Print_Debug_Info;

   use Ada.Streams, System.Garlic.Protocols, System.Garlic.Types;



--  =============================
   Prio_Ceiling : Integer := 29;

   Num_Send : Integer := 0;
   Num_Recv : Integer := 0;

   --  =============================

   subtype Error_Type is Exceptions.Error_Type;


   Activated   : Boolean := False;
   Initialized : Boolean := False;

--   Null_Channel : Integer := 0;
--   Null_Mac_Addr : Integer := 0;
   --  There should not be any concurrent access on this table.

   Max_Incoming   : constant := 5;
   Null_Incoming  : constant := 0;
   First_Incoming : constant := 1;
   Incomings      : array (First_Incoming .. Max_Incoming) of LOCAL_Info;
   Last_Incoming  : Natural := Null_Incoming;

   --  =============================  -- jl MODIFICO
   Incomings_Get_Data : array (1 .. Max_Incoming) of String (1 .. 19);
   --  =============================

   Allocate_Acceptor_Task  : Allocate_Acceptor_Procedure;
--   Allocate_Connector_Task : Allocate_Connector_Procedure;  --  jl REV

   procedure Receive_One_Stream
     (Peer  : in LOCAL_Info;
      PID   : in out Partition_ID;
      Error : in out Error_Type);


   procedure Receive
     (Peer : in LOCAL_Info;
      Data  : access Stream_Element_Array;
      Length : out Stream_Element_Count;
      Error : in out Error_Type);
   pragma Inline (Receive);
   pragma Export (Ada, Receive, "GLADE_Physical_Receive");

   procedure Send
     (Peer  : in LOCAL_Info;
      Data  : access Stream_Element_Array;
      Prio  : in Natural;  --  jl
      From  : in Stream_Element_Count;
      Error : in out Error_Type);

   pragma Inline (Send);
   pragma Export (Ada, Send, "GLADE_Physical_Send");
   --  Receive and send data. Receive loops as long as Data has not
   --  been filled and Send as long as everything has not
   --  been sent. These two procedures are exported so that they can
   --  be used in profiling programs without being in the spec.

   Self_Reference   : Protocol_Access;

--  ------------------------------------------------
--  MIO
--  ------------------------------------------------
   function Value (Mac_Channel : String) return LOCAL_Info;
--  ------------------------------------------------
   -------------------------
   -- Accept_Until_Closed --
   -------------------------
   procedure Accept_Until_Closed
     (Incoming : in Natural) is
   begin
      loop
         declare
            Peer : LOCAL_Info;
            MiPID  : Partition_ID;
         begin
            Peer := Incomings (Incoming);
            Put_Line ("-- Accept_Until_Closed: Data_Banner --");
            MiPID := Null_PID;
            Receive_Until_Closed (Peer, MiPID);
         end;
      end loop;
   end Accept_Until_Closed;

   --------------
   -- Activate --
   --------------
   procedure Activate
     (Protocol : access DLOCAL_Protocol;
      Error    : in out Error_Type)
   is
   begin
      Put_Line ("--------I- Activate-----------");
      if Activated then
         return;
      end if;

      Activated := True;

      if Options.Is_Pure_Client then
         return;
      end if;

      for I in First_Incoming .. Last_Incoming loop
         Allocate_Acceptor_Task (I);
         null;
      end loop;
      Put_Line ("Protocol Local activated");
      Put_Line ("--------F- Activate-----------");
   end Activate;

   ------------
   -- Create --
   ------------

   function Create return Protocol_Access is
   begin
      if Self_Reference /= null then
         return null;
      end if;
      Self_Reference := new DLOCAL_Protocol;
      return Self_Reference;
   end Create;


   --------------
   -- Get_Data --
   --------------

   function Get_Data
     (Protocol : access DLOCAL_Protocol)
     return String_Array_Access
   is
      Result : String_Array_Access;
   begin
      if Options.Is_Pure_Client
        or else Last_Incoming = Null_Incoming
      then
         return null;
      end if;

      Result := new String_Array (First_Incoming .. Last_Incoming);
      for I in Result'Range loop
         Result (I) := new String'(Incomings_Get_Data (I));
      end loop;
      return Result;
   end Get_Data;

   --------------
   -- Get_Name --
   --------------

   function Get_Name
     (Protocol : access DLOCAL_Protocol)
     return String is
   begin
      return "local";
   end Get_Name;

   ----------------
   -- Initialize --
   ----------------
   --  LLega una vez mas de lo necesario y se sale por retur de not REQUIRED

   procedure Initialize
     (Protocol  : access DLOCAL_Protocol;
      Self_Data : in String;
      Required  : in Boolean;
      Performed : out Boolean;
      Error     : in out Error_Type)
   is
      Self  : LOCAL_Info;  --   := Null_LOCAL_Info;
      Index : Natural;
   begin
      Put_Line ("--------I- Initialize-----------");
      Put_Line ("Try to initialize RT-EP with self data """
                  & Self_Data & """");

      --  If this new data is not really needed and if there are
      --  already other data available then skip this one.

      if not Required
        and then Last_Incoming /= Null_Incoming
      then
         pragma Debug (D ("Initialization not required and RT-EP empty"));
         Put_Line ("Initialization not required and RT-EP empty");
         Performed := False;
         Put_Line ("--------F- Initialize-----------");
         return;
      end if;

      --  If anything goes wrong then the registration is not
      --  performed.

      Performed := False;

      if not Initialized then
         pragma Debug (D ("Initialize RT-EP"));
         Put_Line ("Initialize RT-EP");
         --  Inicializo la tabla que relaciona los Identificadores
         --  con sus dir. Mac
         --   Tam_Tabla_Id_Mac := Rellena_Tabla_Id_Mac;
         begin
            Put_Line ("Init_Comm");
            if (Local_Protocol.Init_Local_Protocol (Prio_Ceiling) /= 0) then
               Put_Line ("Init_Comm error");
               return;
            end if;
         exception when others =>
            Put_Line ("Error inicializacion RT-EP");
            return;
         end;
         Initialized := True;
      end if;

      if Self_Data'Length /= 0 then
         Self := Value (Self_Data);
      else
         null;
      end if;

      if Options.Is_Pure_Client then
         return;
      end if;

      for I in First_Incoming .. Last_Incoming loop
         if Incomings (I).Channel = Self.Channel then
            Put_Line ("Use already channel " & Self.Channel'Img);
            return;
         end if;
      end loop;

      Last_Incoming := Last_Incoming + 1;
      Index := Last_Incoming;
      Incomings (Index) := Self;
      Performed := True;
      Put_Line ("Hasta_Aqui_LLEGA_5");

      Incomings_Get_Data (Index) := Self_Data;

      Put_Line ("New self data """ & Self_Data & """ accepted");
      Put_Line ("--------F- Initialize-----------");

   end Initialize;

   -------------
   -- Receive --
   -------------
   procedure Receive
     (Peer  : in LOCAL_Info;
      Data  : access Stream_Element_Array;
      Length : out Stream_Element_Count;
      Error : in out Error_Type)
   is
--      Peer_Variable : LOCAL_Info := Peer;
      Mi_Length : Integer;  -- Unsigned_16;
      Prio_Recv : Integer;
   begin
      begin
         Num_Recv := Num_Recv + 1;
--  jln  Put_Line ("xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx");
         Put_Line ("*** jl Receive -> LOCAL(RT-EP)::"&Num_Recv'Img&":: <-");
--         Put_Line ("*** Mac: " & Peer.Mac_Addr'Img);
--  jln         Put_Line ("*** Canal :" & Peer.Channel'Img);


         Local_Protocol.Recv_Info (1,
                     Peer.Channel,
                     Data,
                     Mi_Length,
                     Prio_Recv);

--         Recv_RTEP (Peer_Variable.Mac_Addr, Peer_Variable.Channel,
--                       Data (1 .. 1492), Mi_Length, Data_Priority_Recv);

         Length := Stream_Element_Count (Mi_Length);
         Put_Line ("*-*-*-*- RECIBIDOOOOO ::"&Num_Recv'Img&"::-*-*-*-*");
--  jln Put_Line ("xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx");
--         Put_Line ("Longitud recibida: " & Mi_Length'Img);
      exception when others =>
         Put_Line ("Receive RTEP mal");
      end;
   end Receive;

   -------------
   -- Receive --
   -------------

   function Receive
     (Protocol  : access DLOCAL_Protocol;
      Timeout   : Duration)
     return Boolean
   is
   begin
      Put_Line ("%%%%%%%%%%%%%%% RECEIVE no debia %%%%%%%%");
      return False;
   end Receive;

   ------------------------
   -- Receive_One_Stream --
   ------------------------
   procedure Receive_One_Stream
     (Peer  : in LOCAL_Info;
      PID   : in out Partition_ID;
      Error : in out Error_Type)
   is
      Old_PID    : Partition_ID;
      Length     : Stream_Element_Count;
      Filtered   : Stream_Element_Access;
      Received   : Stream_Element_Access;
      Unfiltered : Stream_Element_Access;
      Opcode     : Any_Opcode;

   begin
      --  jl LOCAL_PROTOCOL
      if Local_Protocol.Is_Initialized = False then
         Local_Protocol.Init_Mutex_Sem;
      end if;

--  jlReceived := new Stream_Element_Array (1 .. Rt_Ep_Protocol.Max_Info_Size);
      Received := new Stream_Element_Array (1 .. 1492);
      Receive (Peer, Received, Length, Error);
      if Found (Error) then
         return;
      end if;

      Filtered := new Stream_Element_Array (1 .. Length);
      Filtered (1 .. Length) := Received (1 .. Length);
      if Received /= null then
         Free (Received);
      else
         Put_Line ("******************");
         Put_Line ("Received no borrado");
         Put_Line ("******************");
      end if;

      Old_PID := PID;  --  Null_Pid si viene de Accept_Until_Closed
      Analyze_Stream (PID, Self_Reference, Opcode,
                      Unfiltered, Filtered, 0, Error);
      if Found (Error) then
         return;
      end if;

      if Old_PID /= PID then
         Soft_Links.Activity_Detected;
         if Old_PID = Null_PID then
            Set_Online (PID, True);
         else
            Set_Online (Old_PID, False);
            Set_Online (PID, True);
         end if;
      end if;

      if Filtered /= null then
         Free (Filtered);
      else
         Put_Line ("******************");
         Put_Line ("Filtered no borrado");
         Put_Line ("******************");
      end if;

      Process_Stream (PID, Opcode, Unfiltered, Error);

      Put_Line ("Pasado Process stream");

      if Unfiltered /= null then
         Free (Unfiltered);
      else
         Put_Line ("******************");
         Put_Line ("Unfiltered no borrado");
         Put_Line ("******************");
      end if;

   end Receive_One_Stream;

   --------------------------
   -- Receive_Until_Closed --
   --------------------------

   procedure Receive_Until_Closed
     (Peer : in LOCAL_Info;
      PID  : in out Partition_ID)

   is
      Error      : Error_Type;
   begin
      while not Found (Error) loop
         Receive_One_Stream (Peer, PID, Error);
      end loop;
      Catch (Error);

      --  If this connection is broken before partition
      --  identification, then try to rescue PID especially for the
      --  boot server. ???: When does this happen?

      if PID = Null_PID then
         PID := Boot_PID;
      end if;

      --  Protect against multiple attempts to close socket and
      --  prevent any other task from using Incomings (N).Socket
      --  value. The issue is that a task may keep a copy of this file
      --  descriptor. In the meantime, if this file descriptor is
      --  closed and then reallocated for a connection with another
      --  connection, the task will use its copy to communicate with
      --  the wrong partition.

      Set_Online (PID, False);

      --  Signal to the heart that we got an error on this partition

      pragma Debug
        (D ("Notify error on pid" & PID'Img));

      Notify_Partition_Error (PID);
   end Receive_Until_Closed;

   ------------------------
   -- Register_Task_Pool --
   ------------------------

   procedure Register_Task_Pool
     (Allocate_Acceptor  : in Allocate_Acceptor_Procedure) is
   begin
      Allocate_Acceptor_Task  := Allocate_Acceptor;
   end Register_Task_Pool;

   ----------
   -- Send --
   ----------

   procedure Send
     (Protocol  : access DLOCAL_Protocol;
      Partition : in Partition_ID;
      Data      : access Stream_Element_Array;
      Prio_Stream : in Natural;  --  jl
      Error     : in out Error_Type)
   is
      Info     : LOCAL_Info;
      First    : Stream_Element_Count := Data'First + Unused_Space;

   begin

      Info := Value (Read_Partition_Links_OUT_LOC_DATA   --  MODIFICO
                     (Integer (Partition)));             --  MODIFICO

      --  jl LOCAL_PROTOCOL
      if Local_Protocol.Is_Initialized = False then
         Local_Protocol.Init_Mutex_Sem;
      end if;



      Put_Line ("---SENDDDDDDDDD---");
--      if Info.Mac_Addr = Null_Mac_Addr then
--         Throw (Error, "Send: Cannot connect with peer without location");
--         return;
--      end if;

      Set_Online (Partition, True);


      Put_Line ("-- Send --");


      Send (Info, Data, Prio_Stream, First, Error);  --  jl


      --      Send (Info, Data, 3, First, Error);  --  jl
      if Found (Error) then
         --  Protect against multiple attempts to close socket and
         --  prevent any other task from using Incomings (N).Socket
         --  value. The issue is that a task may keep a copy of this file
         --  descriptor. In the meantime, if this file descriptor is
         --  closed and then reallocated for a connection with another
         --  connection, the task will use its copy to communicate with
         --  the wrong partition.

         Set_Online (Partition, False);
      end if;

   end Send;

   ----------
   -- Send --
   ----------
   procedure Send
     (Peer  : in LOCAL_Info;
      Data  : access Stream_Element_Array;
      Prio  : in Natural;
      From  : in Stream_Element_Count;
      Error : in out Error_Type)

   is
      First : Ada.Streams.Stream_Element_Offset := From;
      Last  : constant Ada.Streams.Stream_Element_Offset := Data'Last;

      --  Porque tiene que ser una variable la direccion MAC
--      Peer_Variable : LOCAL_Info := Peer;

      Mi_Var : Ada.Streams.Stream_Element_Offset := Last - First + 1;
--      Dato_Envio : Stream_Element_Access;

   begin
      begin
         Num_Send := Num_Send + 1;
--  jln  Put_Line ("xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx");
         Put_Line ("*** jl Send -> LOCAL (RT-EP)::"&Num_Send'Img&":: <-");
--         Put_Line ("*** Mac: " & Peer.Mac_Addr'Img);
--  jln    Put_Line ("*** Canal :" & Peer.Channel'Img);

--         Dato_Envio := new Stream_Element_Array (1 .. 1492);
--         Dato_Envio (1 .. Mi_Var) := Data (First .. Last);


         --  jl LOCAL_PROTOCOL
         Local_Protocol.Send_Info (1,
                                   Peer.Channel,
                                   Data,
                                   Integer (Mi_Var),
                                   To_Local_Protocol_Priority (
                               Global_Priorities.Priority (Prio)));

--         Send_RTEP (Peer_Variable.Mac_Addr, Peer_Variable.Channel,
--                    Dato_Envio (1 .. 1492),
--                    Unsigned_16 (Mi_Var),
--                    To_Rt_Ep_Priority (Global_Priorities.Priority (Prio)));

--         if Dato_Envio /= null then
--            Free (Dato_Envio);
--         end if;

--  jln         Put_Line ("***  Longitud(Last-First) = " & Mi_Var'Img);
--  jln         Put_Line ("***  Longitud(Length)     = " & Data'Length'Img);
--  jln         Put_Line ("***  Longitud(Size)     = " & Data'Size'Img);
         Put_Line ("xxxxxxxxx Enviado ::"&Num_Send'Img&"::xxxxxxxxxxxxxx");
      exception when others =>
         Put_Line ("Error send LOCAL");
         null;
      end;
   end Send;

   -------------------
   -- Set_Boot_Data --
   -------------------

   procedure Set_Boot_Data
     (Protocol  : access DLOCAL_Protocol;
      Boot_Data : in String;
      Error     : in out Error_Type) is
   begin

      if not Initialized then
         pragma Debug (D ("Initialize RT-EP"));
         Put_Line ("Initialize RT-EP");
         --  Inicializo la tabla que relaciona los Identificadores
         --  con sus dir. Mac
--         Tam_Tabla_Id_Mac := Rellena_Tabla_Id_Mac;
         begin
            Put_Line ("*** Init_Comm Set_Boot_Data");
            if (Local_Protocol.Init_Local_Protocol (Prio_Ceiling) /= 0) then
               Put_Line ("*** ERROR : Init_Comm Set_Boot_Data");
               return;
            end if;
            delay (3.0);
         exception when others =>
            Put_Line ("Error inicializacion LOCAL");
            return;
         end;
         Initialized := True;
      end if;

   end Set_Boot_Data;

   --------------
   -- Shutdown --
   --------------

   procedure Shutdown
     (Protocol : access DLOCAL_Protocol)
   is
   begin
      if not Initialized then
         return;
      end if;

      Put_Line (" Shutdown realizado");
   end Shutdown;


   -----------
   -- Value --
   -----------
   function Value (Mac_Channel : String) return LOCAL_Info is
      Mi_Info : LOCAL_Info;
   begin
      for J in reverse Mac_Channel'Range loop
         if Mac_Channel (J) = ':' then
            --  Put_Line ("J= "& J'Img);
            declare
--               Mac_Addr_TEXT : constant String := Mac_Channel
--                                        (Mac_Channel'First .. J - 1);
--               Mac_Addr : Integer := Null_Mac_Addr;
--               Channel : Integer := 0;

            begin
               --  Pasar mac address
--               Mac_Addr := 2;  --  Convierte_Mac_Addr (Mac_Addr_TEXT);

               Mi_Info.Channel :=  Integer'Value (Mac_Channel
                                 (J +  1 .. Mac_Channel'Last));
               --  RT_EP_Channel_Number'Value (Mac_Channel
                              --   (J +  1 .. Mac_Channel'Last));
               --  Put_Line ("Antes del return del value");
--               return (Mac_Addr, Channel);
               return Mi_Info;
            end;
         end if;
      end loop;
      Mi_Info.Channel := 0;
      Put_Line ("Error al introducir direcciones");
      return Mi_Info;
   end Value;

end System.Garlic.Protocols.Local;
