------------------------------------------------------------------------------
--                                                                          --
--                            GLADE COMPONENTS                              --
--                                                                          --
--          S Y S T E M . G A R L I C . P R O T O C O L S . R T _ E P       --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                            $Revision: 1.6 $
--                                                                          --
--         Copyright (C) 1996-2000 Free Software Foundation, Inc.           --
--                                                                          --
-- GARLIC is free software;  you can redistribute it and/or modify it under --
-- terms of the  GNU General Public License  as published by the Free Soft- --
-- ware Foundation;  either version 2,  or (at your option)  any later ver- --
-- sion.  GARLIC is distributed  in the hope that  it will be  useful,  but --
-- WITHOUT ANY WARRANTY;  without even the implied warranty of MERCHANTABI- --
-- LITY or  FITNESS FOR A PARTICULAR PURPOSE.  See the  GNU General Public  --
-- License  for more details.  You should have received  a copy of the GNU  --
-- General Public License  distributed with GARLIC;  see file COPYING.  If  --
-- not, write to the Free Software Foundation, 59 Temple Place - Suite 330, --
-- Boston, MA 02111-1307, USA.                                              --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable  to  be --
-- covered  by the  GNU  General  Public  License.  This exception does not --
-- however invalidate  any other reasons why  the executable file  might be --
-- covered by the  GNU Public License.                                      --
--                                                                          --
--               GLADE  is maintained by ACT Europe.                        --
--               (email: glade-report@act-europe.fr)                        --
--                                                                          --
------------------------------------------------------------------------------
with Text_IO; use Text_IO;


with Global_Priorities.Mapping;
use Global_Priorities.Mapping;



--  Para Comparacion canales U_16
with Interfaces;                          use Interfaces;

with Rt_Ep_Protocol;                      use Rt_Ep_Protocol;

--  with GNAT.OS_Lib;                         use GNAT.OS_Lib;
--  with GNAT.Sockets;                        use GNAT.Sockets;

--  with Ada.Exceptions;                      --  use Ada.Exceptions;

with System.Garlic.Debug;                 use System.Garlic.Debug;
with System.Garlic.Exceptions;            use System.Garlic.Exceptions;
with System.Garlic.Heart;                 use System.Garlic.Heart;
with System.Garlic.Options;
with System.Garlic.Partitions;            use System.Garlic.Partitions;
--  jl MODIFICO with System.Garlic.Physical_Location;
--  jl MODIFICO use System.Garlic.Physical_Location;
--  with System.Garlic.Platform_Specific;
with System.Garlic.Protocols;             use System.Garlic.Protocols;
with System.Garlic.Soft_Links;
with System.Garlic.Streams;               use System.Garlic.Streams;
--  with System.Garlic.Table;
with System.Garlic.Types;                 use System.Garlic.Types;
with System.Garlic.Utils;                 use System.Garlic.Utils;

with System.Storage_Elements;             use System.Storage_Elements;

package body System.Garlic.Protocols.Rt_Ep is

   Private_Debug_Key : constant Debug_Key :=
     Debug_Initialize ("S_GAPRRT", "(s-gaprrt): ");

   procedure D
     (Message : in String;
      Key     : in Debug_Key := Private_Debug_Key)
     renames Print_Debug_Info;

   use Ada.Streams, System.Garlic.Protocols, System.Garlic.Types;

   --  =============================
   Prio_Ceiling : Integer := 30;
--   Host_Mac_Addr : Ethernet_Address := (16#00#, 16#01#, 16#02#,
--                                        16#08#, 16#14#, 16#1f#);
--   Host_Default_Channel : RT_EP_Channel_Number := 1;
--   Data_Priority_Send : Rt_Ep_Protocol.RT_EP_Priority := 2;
   Data_Priority_Recv : Rt_Ep_Protocol.RT_EP_Priority := 2;
--   Host_Name : String (1 .. 2) := "c1";


   ----------------------
   -- Mis tipos (VALUE)--
   ----------------------
   type IdMac_MacAddr_Type is record
      Id : String (1 .. 7);
      Mac : Ethernet_Address;
   end record;

   Max_Tam_Tabla_Id_Mac : constant Integer := 4;
   Tam_Tabla_Id_Mac : Integer := 0;

   Tabla_Id_Mac : array (1 .. Max_Tam_Tabla_Id_Mac) of IdMac_MacAddr_Type;


   --  =============================

   subtype Error_Type is Exceptions.Error_Type;

   --  Initialize can be invoked several times to register several
   --  self locations. But some initializations of this unit should
   --  happen only once. Initialize has also to be invoked at least
   --  once.

   Activated   : Boolean := False;
   Initialized : Boolean := False;

   Null_Channel : RT_EP_Channel_Number := 1;
   Null_Mac_Addr : Ethernet_Address := (16#01#, 16#01#, 16#01#,
                                        16#01#, 16#01#, 16#01#);
   Null_RTEP_Info : constant RTEP_Info
     := (Null_Mac_Addr, Null_Channel);


   --     --  We need a table to maintain all the outgoing connections. In case
   --     --  of Shutdown and concurrent connections, we may have concurrent
   --     --  accesses. We need a special table for this.

--   package Outgoings is
--      new System.Garlic.Table.Medium
--     (Partition_ID,
--      Null_PID,
--      First_PID,
--      Partition_ID_Increment,
--      Partition_ID_Increment,
--      RTEP_Info,
--      Null_RTEP_Info);


   --  There should not be any concurrent access on this table.

   Max_Incoming   : constant := 5;
   Null_Incoming  : constant := 0;
   First_Incoming : constant := 1;
   Incomings      : array (First_Incoming .. Max_Incoming) of RTEP_Info;
   Last_Incoming  : Natural := Null_Incoming;

   --  =============================  -- jl MODIFICO
   Incomings_Get_Data : array (1 .. Max_Incoming) of String (1 .. 19);
   --  =============================

   Allocate_Acceptor_Task  : Allocate_Acceptor_Procedure;
--   Allocate_Connector_Task : Allocate_Connector_Procedure;  --  jl REV

   procedure Receive_One_Stream
     (Peer  : in RTEP_Info;
      PID   : in out Partition_ID;
      Error : in out Error_Type);


   procedure Receive
     (Peer : in RTEP_Info;
      Data  : access Stream_Element_Array;
      Length : out Stream_Element_Count;
      Error : in out Error_Type);
   pragma Inline (Receive);
   pragma Export (Ada, Receive, "GLADE_Physical_Receive");

   procedure Send
     (Peer  : in RTEP_Info;
      Data  : access Stream_Element_Array;
      Prio  : in Natural;  --  jl
      From  : in Stream_Element_Count;
      Error : in out Error_Type);

   pragma Inline (Send);
   pragma Export (Ada, Send, "GLADE_Physical_Send");
   --  Receive and send data. Receive loops as long as Data has not
   --  been filled and Send as long as everything has not
   --  been sent. These two procedures are exported so that they can
   --  be used in profiling programs without being in the spec.

   Self_Reference   : Protocol_Access;

--  ------------------------------------------------
--  MIO
--  ------------------------------------------------
   subtype RTEP_Data_Type is
          Stream_Element_Array (1 .. Rt_Ep_Protocol.Max_Info_Size);
   procedure Send_RTEP is new Generic_Send_Info
     (Data_Type => RTEP_Data_Type);
   procedure Recv_RTEP is new Generic_Recv_Info
     (Data_Type => RTEP_Data_Type);

   function Rellena_Tabla_Id_Mac return Integer;
--   function Busca_Mac_Addr (Id_Mac : String) return Ethernet_Address;
   function Convierte_Mac_Addr (Id_Mac : String) return Ethernet_Address;
   function Value (Mac_Channel : String) return RTEP_Info;
--  ------------------------------------------------
   -------------------------
   -- Accept_Until_Closed --
   -------------------------
   procedure Accept_Until_Closed
     (Incoming : in Natural) is
   begin
      loop
         declare
            Peer : RTEP_Info;
            MiPID  : Partition_ID;
         begin
            Peer := Incomings (Incoming);
--           Put_Line ("-- Accept_Until_Closed: Data_Banner --");
            MiPID := Null_PID;
            Receive_Until_Closed (Peer, MiPID);
         end;
      end loop;
   end Accept_Until_Closed;

   --------------
   -- Activate --
   --------------
   procedure Activate
     (Protocol : access RTEP_Protocol;
      Error    : in out Error_Type)
   is
   begin
--      Put_Line ("--------I- Activate-----------");
      if Activated then
         return;
      end if;

      Activated := True;

      if Options.Is_Pure_Client then
         return;
      end if;

      for I in First_Incoming .. Last_Incoming loop
         Allocate_Acceptor_Task (I);
      end loop;
--      Put_Line ("Protocol rtep activated");
--      Put_Line ("--------F- Activate-----------");
   end Activate;

   ------------
   -- Create --
   ------------

   function Create return Protocol_Access is
   begin
      if Self_Reference /= null then
         return null;
      end if;
      Self_Reference := new RTEP_Protocol;
      return Self_Reference;
   end Create;


   --------------
   -- Get_Data --
   --------------

   function Get_Data
     (Protocol : access RTEP_Protocol)
     return String_Array_Access
   is
      Result : String_Array_Access;
   begin
      if Options.Is_Pure_Client
        or else Last_Incoming = Null_Incoming
      then
         return null;
      end if;
      Result := new String_Array (First_Incoming .. Last_Incoming);
      for I in Result'Range loop
         Result (I) := new String'(Incomings_Get_Data (I));
      end loop;
      return Result;
   end Get_Data;

   --------------
   -- Get_Name --
   --------------

   function Get_Name
     (Protocol : access RTEP_Protocol)
     return String is
   begin
      return "rt_ep";
   end Get_Name;

   ----------------
   -- Initialize --
   ----------------
   --  LLega una vez mas de lo necesario y se sale por retur de not REQUIRED

   procedure Initialize
     (Protocol  : access RTEP_Protocol;
      Self_Data : in String;
      Required  : in Boolean;
      Performed : out Boolean;
      Error     : in out Error_Type)
   is
--    Host  : constant RTEP_Info := Value (Host_Name);
      Self  : RTEP_Info := Null_RTEP_Info;
      Index : Natural;
   begin
--      Put_Line ("--------I- Initialize-----------");
--      Put_Line ("Try to initialize RT-EP with self data """
--                  & Self_Data & """");

      --  If this new data is not really needed and if there are
      --  already other data available then skip this one.

      if not Required
        and then Last_Incoming /= Null_Incoming
      then
         pragma Debug (D ("Initialization not required and RT-EP empty"));
         Put_Line ("Initialization not required and RT-EP empty");
         Performed := False;
--         Put_Line ("--------F- Initialize-----------");
         return;
      end if;

      --  If anything goes wrong then the registration is not
      --  performed.

      Performed := False;

      if not Initialized then
         pragma Debug (D ("Initialize RT-EP"));
--         Put_Line ("Initialize RT-EP");
         --  Inicializo la tabla que relaciona los Identificadores
         --  con sus dir. Mac
         Tam_Tabla_Id_Mac := Rellena_Tabla_Id_Mac;
--         Outgoings.Initialize;
         begin
--            Put_Line ("Init_Comm");
            if (Rt_Ep_Protocol.Init_Comm (Prio_Ceiling) /= 0) then
               Put_Line ("Init_Comm error");
               return;
            end if;
         exception when Creation_Error | Unexpected_Error =>
            Put_Line ("Error inicializacion RT-EP");
            return;
         end;
         Initialized := True;
      end if;
--      Put_Line ("Hasta_Aqui_LLEGA_1");

      if Self_Data'Length /= 0 then
         Self := Value (Self_Data);
         --  When there is a self location to bind on, check that this
         --  location concerns the current host. Otherwise, there is a
         --  real problem.
--       Put_Line ("Self : " & Self.Mac_Addr (1)'Img &
--                 " Host : " & Host.Mac_Addr (1)'Img);
--       Put_Line ("Self : " & Self.Mac_Addr (2)'Img &
--                 " Host : " & Host.Mac_Addr (2)'Img);
--       Put_Line ("Self : " & Self.Mac_Addr (3)'Img &
--                 " Host : " & Host.Mac_Addr (3)'Img);
--       Put_Line ("Self : " & Self.Mac_Addr (4)'Img &
--                 " Host : " & Host.Mac_Addr (4)'Img);
--       Put_Line ("Self : " & Self.Mac_Addr (5)'Img &
--                 " Host : " & Host.Mac_Addr (5)'Img);
--       Put_Line ("Self : " & Self.Mac_Addr (6)'Img &
--                 " Host : " & Host.Mac_Addr (6)'Img);

--         if Self.Mac_Addr /= Host_Mac_Addr then
--            Put_Line ("No es una direccion de este host");
--            Ada.Exceptions.Raise_Exception
--              (Program_Error'Identity, "Incorrect tcp self location: "
--                 & Self_Data);
--         end if;
      else
--         Put_Line ("Hasta_Aqui_LLEGA_2");
--         Self := Host;
--         Put_Line ("Initialize error");
--         Ada.Exceptions.Raise_Exception
--           (Program_Error'Identity, "Incorrect tcp self location: "
--              & Self_Data);
         null;
      end if;
--      Put_Line ("Hasta_Aqui_LLEGA_3");

      if Options.Is_Pure_Client then
         return;
      end if;
--      Put_Line ("Hasta_Aqui_LLEGA_4");

--      Put_Line ("Hasta aqui bien");
--      Ada.Exceptions.Raise_Exception
--              (Program_Error'Identity, "Incorrect tcp self location: "
--                 & Self_Data);

      --  Skip this incoming location if it does already exist.



      for I in First_Incoming .. Last_Incoming loop
         if Incomings (I).Channel = Self.Channel then
            Put_Line ("Use already channel " & Self.Channel'Img);
            return;
         end if;
      end loop;

      Last_Incoming := Last_Incoming + 1;
      Index := Last_Incoming;
      Incomings (Index) := Self;
      Performed := True;
--      Put_Line ("Hasta_Aqui_LLEGA_5");

--  jl      Incomings_Get_Data (Index) := Self_Data;
      Incomings_Get_Data (Index) := Self_Data;

--      Put_Line ("New self data """ & Self_Data & """ accepted");
--      Put_Line ("--------F- Initialize-----------");

   end Initialize;

   -------------
   -- Receive --
   -------------
   procedure Receive
     (Peer  : in RTEP_Info;
      Data  : access Stream_Element_Array;
      Length : out Stream_Element_Count;
      Error : in out Error_Type)
   is
--      First : Ada.Streams.Stream_Element_Offset := Data'First;
--      Last  : constant Ada.Streams.Stream_Element_Offset := Data'Last;
      --  Porque tiene que ser una variable la direccion MAC
      Peer_Variable : RTEP_Info := Peer;
      Mi_Length : Unsigned_16;
   begin
      begin
--       Put_Line ("xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx");
--         Put_Line ("*** jl Receive -> RT-EP(RT-EP) <-");
--         Put_Line ("*** Mac: "
--                  & Peer.Mac_Addr (1)'Img & ":" & Peer.Mac_Addr (2)'Img
--                  & ":" & Peer.Mac_Addr (3)'Img & ":" & Peer.Mac_Addr (4)'Img
--                & ":" & Peer.Mac_Addr (5)'Img & ":" & Peer.Mac_Addr (6)'Img);
--         Put_Line ("*** Canal :" & Peer.Channel'Img);


         Recv_RTEP (Peer_Variable.Mac_Addr, Peer_Variable.Channel,
                       Data (1 .. 1492), Mi_Length, Data_Priority_Recv);
         Length := Stream_Element_Count (Mi_Length);
--         Put_Line ("*-*-*-*- RECIBIDOOOOO -*-*-*-*");
--       Put_Line ("xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx");
--         Put_Line ("Longitud recibida: " & Mi_Length'Img);
      exception when Invalid_Channel | Synch_Error | Unexpected_Error =>
         Put_Line ("Receive RTEP mal");
      end;
   end Receive;

   -------------
   -- Receive --
   -------------

   function Receive
     (Protocol  : access RTEP_Protocol;
      Timeout   : Duration)
     return Boolean
   is
      --        RSet     : Socket_Set_Type;
      --        WSet     : Socket_Set_Type;
      --        Info     : Socket_Info;
      --        Done     : Boolean := False;
      --        Error    : Error_Type;
      --        PID      : Partition_ID;
      --        Selector : Selector_Type;
      --        Status   : Selector_Status;

   begin
      --        --  This procedure is called in a No_Tasking env. No need to
      --        --  protect against concurrent access.

      --        Put_Line ("-- El Receive --");

      --        Empty (RSet);
      --        Empty (WSet);

      --        for O in First_PID .. Outgoings.Last loop
      --           Info := Outgoings.Get_Component (O);
      --           if Info.Socket /= No_Socket then
      --      pragma Debug (D ("include socket " & Image (Info.Socket)));

      --              Set (RSet, Info.Socket);
      --           end if;
      --        end loop;

      --        if Is_Empty (RSet) then
      --           return True;
      --        end if;

      --        Create_Selector (Selector);
      --        Check_Selector  (Selector, RSet, WSet, Status, Timeout);
      --        Close_Selector  (Selector);
      --        pragma Debug (D ("select returned with status " & Status'Img));

      --        if Status = Expired then
      --           Empty (RSet);
      --           Empty (WSet);
      --           return False;
      --        end if;

      --        for O in First_PID .. Outgoings.Last loop
      --           Info := Outgoings.Get_Component (O);
      --           if Info.Socket /= No_Socket
      --             and then Is_Set (RSet, Info.Socket)
      --           then
      --        pragma Debug (D ("something ready from partition" & O'Img));

      --              Done := True;
      --              PID  := O;
      --              Receive_One_Stream (Info.Socket, PID, Error);
      --              if Found (Error) then
      --                 Close_Socket_Ignoring_Error (Info.Socket);
      --                 Info.Socket := No_Socket;
      --                 Outgoings.Set_Component (PID, Info);
      --                 Set_Online (PID, False);
      --                 Outgoings.Update;
      --                 Notify_Partition_Error (PID);
      --              end if;
      --           end if;
      --        end loop;

      --        Empty (RSet);
      --        Empty (WSet);
      --        return Done;

      Put_Line ("%%%%%%%%%%%%%%% RECEIVE no debia %%%%%%%%");
      return False;
   end Receive;

   ------------------------
   -- Receive_One_Stream --
   ------------------------
   procedure Receive_One_Stream
     (Peer  : in RTEP_Info;
      PID   : in out Partition_ID;
      Error : in out Error_Type)
   is
      Old_PID    : Partition_ID;
      Length     : Stream_Element_Count;
      Filtered   : Stream_Element_Access;
      Received   : Stream_Element_Access;
      Unfiltered : Stream_Element_Access;
      Opcode     : Any_Opcode;

   begin
      --  pragma Debug (D ("Recv" & Length'Img & " bytes from peer " &
      --                   Image (Peer) & " (pid =" & PID'Img & ")"));

      Received := new Stream_Element_Array (1 .. Rt_Ep_Protocol.Max_Info_Size);
      Receive (Peer, Received, Length, Error);
      if Found (Error) then
         return;
      end if;

--      for I in 1 .. Natural (Length) loop
--         Put (Received (Stream_Element_Offset (I))'Img);
--      end loop;
--      New_Line;


--      Put_Line ("Antes Filtered");
      Filtered := new Stream_Element_Array (1 .. Length);
      Filtered (1 .. Length) := Received (1 .. Length);
--      Put_Line ("Pasado Filtered");
      if Received /= null then
         Free (Received);
      end if;

--      Put_Line (Filtered (1)'Img & Filtered (2)'Img);
--      Put_Line (Filtered (3)'Img & Filtered (4)'Img);
--      Put_Line (Filtered (5)'Img & Filtered (6)'Img);

      Old_PID := PID;  --  Null_Pid si viene de Accept_Until_Closed
      Analyze_Stream (PID, Self_Reference, Opcode,
                      Unfiltered, Filtered, 0, Error);
      if Found (Error) then
         return;
      end if;

--      Put_Line ("Opcode : " & Opcode'Img);

--      Put_Line ("Pasado Analized stream");

      if Old_PID /= PID then
         Soft_Links.Activity_Detected;
         if Old_PID = Null_PID then
            Set_Online (PID, True);
         else
            Set_Online (Old_PID, False);
            Set_Online (PID, True);
         end if;
      end if;

      Process_Stream (PID, Opcode, Unfiltered, Error);

--      Put_Line ("Pasado Process stream");

      if Filtered /= null then
         Free (Filtered);
      end if;
      if Unfiltered /= null then
         Free (Unfiltered);
      end if;
   end Receive_One_Stream;

   --------------------------
   -- Receive_Until_Closed --
   --------------------------

   --  procedure Receive_Until_Closed
   --    (Peer : in Socket_Type;
   --     PID  : in out Partition_ID)
   procedure Receive_Until_Closed
     (Peer : in RTEP_Info;
      PID  : in out Partition_ID)


   is
      Error      : Error_Type;
   begin
      while not Found (Error) loop
         Receive_One_Stream (Peer, PID, Error);
      end loop;
      Catch (Error);

      --  If this connection is broken before partition
      --  identification, then try to rescue PID especially for the
      --  boot server. ???: When does this happen?

      if PID = Null_PID then
         PID := Boot_PID;
      end if;

      --  Protect against multiple attempts to close socket and
      --  prevent any other task from using Incomings (N).Socket
      --  value. The issue is that a task may keep a copy of this file
      --  descriptor. In the meantime, if this file descriptor is
      --  closed and then reallocated for a connection with another
      --  connection, the task will use its copy to communicate with
      --  the wrong partition.

      Set_Online (PID, False);

      --  Signal to the heart that we got an error on this partition

      pragma Debug
        (D ("Notify error on pid" & PID'Img));

      Notify_Partition_Error (PID);
   end Receive_Until_Closed;

   ------------------------
   -- Register_Task_Pool --
   ------------------------

--   procedure Register_Task_Pool  --  jl REV
--     (Allocate_Acceptor  : in Allocate_Acceptor_Procedure;  --  jl REV
--      Allocate_Connector : in Allocate_Connector_Procedure) is  --  jl REV
   procedure Register_Task_Pool
     (Allocate_Acceptor  : in Allocate_Acceptor_Procedure) is
--      Allocate_Connector : in Allocate_Connector_Procedure) is  --  jl REV
   begin
      Allocate_Acceptor_Task  := Allocate_Acceptor;
--      Allocate_Connector_Task := Allocate_Connector;  --  jl REV
   end Register_Task_Pool;

   ----------
   -- Send --
   ----------

   procedure Send
     (Protocol  : access RTEP_Protocol;
      Partition : in Partition_ID;
      Data      : access Stream_Element_Array;
      Prio_Stream : in Natural;  --  jl
      Error     : in out Error_Type)
   is
--      Prio_Stream : Natural := 10;  --  jl NUEVA

      --  Info     : Socket_Info;
      Info     : RTEP_Info;
      First    : Stream_Element_Count := Data'First + Unused_Space;
      --  jl MODIFICO Location : Location_Type;
      --  MiPID    : Partition_ID;


   begin
      --  jl MODIFICO
      --  jl MODIFICO Get_Net_Location (Partition, Location, Error);
      --  jl MODIFICO if Found (Error) then
      --  jl MODIFICO    return;
      --  jl MODIFICO end if;
      --  jl MODIFICO Info := Value (Get_Data (Location));

      Info := Value (Read_Partition_Links_OUT_LOC_DATA   --  MODIFICO
                     (Integer (Partition)));             --  MODIFICO

--      Put_Line ("---SENDDDDDDDDD---");
      if Info.Mac_Addr = Null_Mac_Addr then
         Throw (Error, "Send: Cannot connect with peer without location");
         return;
      end if;

      Set_Online (Partition, True);


--      Put_Line ("-- Send --");


      Send (Info, Data, Prio_Stream, First, Error);  --  jl
--      Send (Info, Data, 3, First, Error);  --  jl
      if Found (Error) then
         --  Protect against multiple attempts to close socket and
         --  prevent any other task from using Incomings (N).Socket
         --  value. The issue is that a task may keep a copy of this file
         --  descriptor. In the meantime, if this file descriptor is
         --  closed and then reallocated for a connection with another
         --  connection, the task will use its copy to communicate with
         --  the wrong partition.

         Set_Online (Partition, False);
      end if;

   end Send;

   ----------
   -- Send --
   ----------
   procedure Send
     (Peer  : in RTEP_Info;
      Data  : access Stream_Element_Array;
      Prio  : in Natural;
      From  : in Stream_Element_Count;
      Error : in out Error_Type)

   is
      First : Ada.Streams.Stream_Element_Offset := From;
      Last  : constant Ada.Streams.Stream_Element_Offset := Data'Last;

      --  Porque tiene que ser una variable la direccion MAC
      Peer_Variable : RTEP_Info := Peer;

--   OJO   Mi_Var : Ada.Streams.Stream_Element_Offset := Last - First;
      Mi_Var : Ada.Streams.Stream_Element_Offset := Last - First + 1;
      Dato_Envio : Stream_Element_Access;

   begin
      begin
--       Put_Line ("xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx");
--         Put_Line ("*** jl Send -> RT-EP (RT-EP) <-");
--         Put_Line ("*** Mac: "
--                  & Peer.Mac_Addr (1)'Img & ":" & Peer.Mac_Addr (2)'Img
--                  & ":" & Peer.Mac_Addr (3)'Img & ":" & Peer.Mac_Addr (4)'Img
--                & ":" & Peer.Mac_Addr (5)'Img & ":" & Peer.Mac_Addr (6)'Img);
--         Put_Line ("*** Canal :" & Peer.Channel'Img);

         Dato_Envio := new Stream_Element_Array (1 .. 1492);
         Dato_Envio (1 .. Mi_Var) := Data (First .. Last);


--         for I in 1 .. Natural (Last) loop
--            Put (Dato_Envio (Stream_Element_Offset (I))'Img);
--         end loop;
--         New_Line;


         Send_RTEP (Peer_Variable.Mac_Addr, Peer_Variable.Channel,
                    Dato_Envio (1 .. 1492),
                    Unsigned_16 (Mi_Var),
                    To_Rt_Ep_Priority (Global_Priorities.Priority (Prio)));
--                    Data_Priority_Send);

         if Dato_Envio /= null then
            Free (Dato_Envio);
         end if;

--         Put_Line ("***  Longitud(Last-First) = " & Mi_Var'Img);
--         Put_Line ("***  Longitud(Length)     = " & Data'Length'Img);
--         Put_Line ("***  Longitud(Size)     = " & Data'Size'Img);
--       Put_Line ("xxxxxxxxxENVIADOxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx");
--      exception when Station_Not_Valid | Station_Not_Found |
--         Invalid_Channel | Synch_Error | Unexpected_Error =>
      exception when others =>
         Put_Line ("Error send RTEP");
         null;
      end;
   end Send;

   -------------------
   -- Set_Boot_Data --
   -------------------

   procedure Set_Boot_Data
     (Protocol  : access RTEP_Protocol;
      Boot_Data : in String;
      Error     : in out Error_Type) is
   begin

      if not Initialized then
         pragma Debug (D ("Initialize RT-EP"));
         Put_Line ("Initialize RT-EP");
         --  Inicializo la tabla que relaciona los Identificadores
         --  con sus dir. Mac
         Tam_Tabla_Id_Mac := Rellena_Tabla_Id_Mac;
--         Outgoings.Initialize;
         begin
--            Put_Line ("*** Init_Comm Set_Boot_Data");
            if (Init_Comm (Prio_Ceiling) /= 0) then
               Put_Line ("*** ERROR : Init_Comm Set_Boot_Data");
               return;
            end if;
            delay (3.0);
         exception when Creation_Error | Unexpected_Error =>
            Put_Line ("Error inicializacion RT-EP");
            return;
         end;
         Initialized := True;
      end if;

   end Set_Boot_Data;

   --------------
   -- Shutdown --
   --------------

   procedure Shutdown
     (Protocol : access RTEP_Protocol)
   is
   begin
      if not Initialized then
         return;
      end if;

      Put_Line (" Shutdown realizado");
   end Shutdown;


   --------------------------
   -- Rellena_Tabla_Id_Mac --
   --------------------------
   function Rellena_Tabla_Id_Mac return Integer is
   begin

      Tabla_Id_Mac (1).Id  := "ctrpc28";
      Tabla_Id_Mac (1).Mac := (16#00#, 16#02#, 16#B3#, 16#D1#, 16#30#, 16#EC#);

      Tabla_Id_Mac (2).Id  := "ctrpc18";
      Tabla_Id_Mac (2).Mac := (16#00#, 16#02#, 16#B3#, 16#D1#, 16#34#, 16#BF#);

      Tabla_Id_Mac (3).Id  := "ctrpcxx";
      Tabla_Id_Mac (3).Mac := (16#00#, 16#01#, 16#02#, 16#08#, 16#14#, 16#1f#);

      return 3; --  devuelvo el numero de elementos
   end Rellena_Tabla_Id_Mac;

   --------------------
   -- Busca_Mac_Addr --
   --------------------
--  jl   function Busca_Mac_Addr (Id_Mac : String) return Ethernet_Address is

--  jl   begin
--  jl      for I in 1 .. Tam_Tabla_Id_Mac loop
--  jl         if Id_Mac = Tabla_Id_Mac (I).Id then
--  jl            return Tabla_Id_Mac (I).Mac;
--  jl         end if;
--  jl      end loop;
--  jl      Put_Line ("Error en Busca_Mac_Addr");
--  jl      return Null_Mac_Addr;
--  jl   end Busca_Mac_Addr;


--     -----------
--     -- Value --
--     -----------
--     function Value (Mac_Channel : String) return RTEP_Info is
--     begin
--        for J in reverse Mac_Channel'Range loop
--           if Mac_Channel (J) = ':' then
--              --  Put_Line ("J= "& J'Img);
--              declare
--  --  jl             Host : constant String := Mac_Channel
--  --  jl                                        (Mac_Channel'First .. J - 1);
--                 Mac_Addr : Ethernet_Address := Null_Mac_Addr;
--                 Channel : RT_EP_Channel_Number := 1;

--              begin
--                 --  Pasar mac address
--  --  jl             Mac_Addr := Busca_Mac_Addr (Host);
--  --  jl             Channel := RT_EP_Channel_Number'Value (Mac_Channel
--  --  jl                                (J +  1 .. Mac_Channel'Last));
--                 return (Mac_Addr, Channel);
--              end;
--           end if;
--        end loop;

--        Put_Line ("Error al introducir direcciones");
--        return Null_RTEP_Info;
--     end Value;

--     --------------------
--     -- Convierte_Mac_Addr --
--     --------------------
--     --  Le Pasas el texto MAC y te obtiene el numero MAC
--     function Convierte_Mac_Addr (Id_Mac : String) return Ethernet_Address is
--        Mac : Ethernet_Address;
--     begin
--        Mac (1) := Interfaces.Unsigned_8'Value (Id_Mac (1 .. 2));
--        Mac (2) := Interfaces.Unsigned_8'Value (Id_Mac (4 .. 5));
--        Mac (3) := Interfaces.Unsigned_8'Value (Id_Mac (7 .. 8));
--        Mac (4) := Interfaces.Unsigned_8'Value (Id_Mac (10 .. 11));
--        Mac (5) := Interfaces.Unsigned_8'Value (Id_Mac (13 .. 14));
--        Mac (6) := Interfaces.Unsigned_8'Value (Id_Mac (16 .. 17));
--        return Mac;
--     end Convierte_Mac_Addr;
   --------------------
   -- Convierte_Mac_Addr --
   --------------------
   --  Le Pasas el texto MAC y te obtiene el numero MAC
   function Convierte_Mac_Addr (Id_Mac : String) return Ethernet_Address is
      Mac : Ethernet_Address;

      function Obten_Valor_Hexadecimal (Texto : String)
                                       return Interfaces.Unsigned_8;

      function Obten_Valor_Hexadecimal (Texto : String)
                                       return Interfaces.Unsigned_8 is
         Suma : Interfaces.Unsigned_8 := 0;
         Basura : String (1 .. 2) := "00";
      begin
         --  Put_Line ("--- ID_MAC ---" & Texto);
         for I in 1 .. 2 loop
            if Texto (I) = 'A' then
               --  Put_Line ("A");
               Suma := Suma + 10;
            elsif Texto (I) = 'B' then
               --  Put_Line ("B");
               Suma := Suma + 11;
            elsif Texto (I) = 'C' then
               --  Put_Line ("C");
               Suma := Suma + 12;
            elsif Texto (I) = 'D' then
               --  Put_Line ("D");
               Suma := Suma + 13;
            elsif Texto (I) = 'E' then
               --  Put_Line ("E");
               Suma := Suma + 14;
            elsif Texto (I) = 'F' then
               --  Put_Line ("F");
               Suma := Suma + 15;
            else
               Basura (2) := Texto (I);
               --  Put_Line ("Basura" & Basura);
               Suma := Suma + Interfaces.Unsigned_8'Value (Basura);
            end if;
            if I = 1 then
               Suma := Suma * 16;
            end if;
         end loop;
         return Suma;
      end Obten_Valor_Hexadecimal;
      Aux : String (1 .. 2);
      Aux2 : String (1 .. 40);  --  jl COSAS RARAS DE LOS STRINGS
   begin
--      Put_Line ("En Convierte_Mac_Addr ->" & Id_Mac);
      Aux2 (1 .. Id_Mac'Length) := Id_Mac;

      Aux := Aux2 (1 .. 2);
--      Aux := Id_Mac (1 .. 2);
      Mac (1) := Obten_Valor_Hexadecimal (Aux);
--      Put_Line ("MAC_1:" & Mac (1)'Img);

      Aux := Aux2 (4 .. 5);
--      Aux := Id_Mac (4 .. 5);
      Mac (2) := Obten_Valor_Hexadecimal (Aux);
--      Put_Line ("MAC_2:" & Mac (2)'Img);

      Aux := Aux2 (7 .. 8);
--      Aux := Id_Mac (7 .. 8);
      Mac (3) := Obten_Valor_Hexadecimal (Aux);
--      Put_Line ("MAC_3:" & Mac (3)'Img);

      Aux := Aux2 (10 .. 11);
--      Aux := Id_Mac (10 .. 11);
      Mac (4) := Obten_Valor_Hexadecimal (Aux);
--      Put_Line ("MAC_4:" & Mac (4)'Img);

      Aux := Aux2 (13 .. 14);
--      Aux := Id_Mac (13 .. 14);
      Mac (5) := Obten_Valor_Hexadecimal (Aux);
--      Put_Line ("MAC_5:" & Mac (5)'Img);

      Aux := Aux2 (16 .. 17);
--      Aux := Id_Mac (16 .. 17);
      Mac (6) := Obten_Valor_Hexadecimal (Aux);
--      Put_Line ("MAC_6:" & Mac (6)'Img);
      return Mac;
   end Convierte_Mac_Addr;

   -----------
   -- Value --
   -----------
   function Value (Mac_Channel : String) return RTEP_Info is
   begin
      for J in reverse Mac_Channel'Range loop
         if Mac_Channel (J) = ':' then
            --  Put_Line ("J= "& J'Img);
            declare
               Mac_Addr_TEXT : constant String := Mac_Channel
                                        (Mac_Channel'First .. J - 1);
               Mac_Addr : Ethernet_Address := Null_Mac_Addr;
               Channel : RT_EP_Channel_Number := 1;

            begin
               --  Pasar mac address
               --  Put_Line ("En Value: (Mac_Addr_TEXT)->" & Mac_Addr_TEXT);
               Mac_Addr := Convierte_Mac_Addr (Mac_Addr_TEXT);
               --  Put_Line ("Tras Convierte_Mac_Addr");
               Channel := RT_EP_Channel_Number'Value (Mac_Channel
                                (J +  1 .. Mac_Channel'Last));
               --  Put_Line ("Antes del return del value");
               return (Mac_Addr, Channel);
            end;
         end if;
      end loop;

      Put_Line ("Error al introducir direcciones");
      return Null_RTEP_Info;
   end Value;

end System.Garlic.Protocols.Rt_Ep;
