/*----------------------------------------------------------------------------
 *-------------------------      M a R T E   O S      ------------------------
 *----------------------------------------------------------------------------
 *                                                             {MARTE_VERSION}
 *
 *                            'e t h e r n e t . h'
 *
 *                                     C
 *
 *                                               
 * File 'ethernet.h'                                                  By Chema.
 *                                                          Jose Maria Martinez
 *                                                         <martinjm@unican.es>
 * Driver independent functions to tx and rx data.
 *
 * Body fuctions on 'nic.c'
 *
 * {MARTE_COPYRIGHT}
 *
 *---------------------------------------------------------------------------*/
/* {<MAR}
 * Chema:
 * Date format DD-MM-YY:
 *
 *
 * 26-11-02: First draft. 
 * 13-01-03: Working version.
 *
 *
 * {MAR>} -------------------------------------------------------------------*/

#ifndef _ETHERNET_H_
#define _ETHERNET_H_

#include <sys/types.h>


/********
  Create
*********/

/* eth_create : This function scans and configure all the Ethernet PCI  */
/*                  ethernet cards in the system. Returns the number of     */
/*                  interfaces found in the station.                        */
/*                  Returns 0 if OK.                                        */

extern ssize_t eth_create(int create_arg);


/******
  Open
*******/

extern ssize_t  eth_open(int file_descriptor, int file_access_mode);

/******
  Read
*******/
/* eth_read : Receives an Ethernet II frame.                            */
/*            The structure of *msg is a Ethernet II frame, as help you */
/*            can use:                                                  */
/*            {                                                         */
/*               struct ethhdr eth_header; // defined in if_ether.h     */
/*               char *info;               // a free reserved area to   */
/*            }                            // store the information.    */
/*            Returns the number of received bytes including the header */
/*            The protocol is in Network byte order !                   */
/*            On error return -1 // pending on a errno implementation   */
/*               on MaRTE*/
/*            The field len is unused.                                  */


extern ssize_t eth_read(int file_descriptor, void *msg, size_t len);

/*******
  Write
********/
/* eth_write : Sends an Ethernet II frame.                               */
/*             *msg - is part of as valid Ethernet frame, it *MUST* have */
/*                    in this order:                                     */
/*            The structure of *msg is as follows:                       */
/*            {                                                          */
/*               struct ethhdr eth_header; // defined in if_ether.h      */
/*               char *info;               // a free reserved area to    */
/*            }                            // store the information.     */
/*            len *MUST* be the size of the ethernet frame.              */
/*            The protocol of ethhdr MUST be in network  byte order.     */
/*            len = ETH_HLEN + Data size.                                */
/*            The protocol is MUST be in network byte order.             */


extern ssize_t eth_write(int file_descriptor, void *msg, size_t len);


/*******
  Close
********/
/* eth_close : Depends on the driver but usually turns off interrupts and  */
/*               stops Tx and Rx engines. */
extern ssize_t eth_close(int file_descriptor);


/*******
  Ioctl
********/

extern ssize_t eth_ioctl(int file_descriptor, ssize_t request, void *priv);


#endif
