----------------------------------------------------------------------------
-------------------------      M a R T E   O S      ------------------------
----------------------------------------------------------------------------
--                                                            {MARTE_VERSION}
--
--                   'e t h e r n e t _ d r i v e r . a d s'
--
--                                     Ada
--
--
-- File 'ethernet_driver.ads'                                         By Chema.
--                                                          Jose Maria Martinez
--                                                         <martinjm@unican.es>
-- Import declarations of eth_ioctl.h
--
--
--{MARTE_COPYRIGHT}
--
-----------------------------------------------------------------------------

with Basic_Integer_Types; use Basic_Integer_Types;

package Ethernet_Driver is

   --  ---------------------------
   --  Ethernet 'Ioctl' Options
   --  ---------------------------

   --  Commands explanation.

   --   Eth_Hardware_Address : Will write in Eth_Addr_Ioctl_Arg the Physical MAC
   --                          address of the current Ethernet card. (the size of
   --                          MAC addr is 48 bits.)

   --  Eth_Blocking_Read : Will turn the read operation blocking. read will not
   --                      return till any data is received.

   --  Eth_Non_Blocking_Read : Will turn the read operation as non blocking.
   --                          The read call will return with or without (in
   --                          case not any) data

   --  Set_Protocol_Filter : Will set the receiving protocol filter. By default
   --                        the driver only receives RT_EP packets. You can
   --                        configure to receive only packets defined in
   --                        Eth_Proto_Arg (unsigned short *) (IP, NETBIOS, etc..
   --                        <see include/drivers/if_ether.h> for protocol
   --                        numbers). If set to 0, the driver will receive all
   --                        protocol packets on the net.

   --  Get_Protocol_Filter : Will write in Eth_Proto_Arg (unsigned short) the current
   --                        protocol filter.
   --                        If protocol 0, the protocol filter is disabled

   type Ioctl_Cmd_No_Args is
     (Eth_Blocking_Read,
      Eth_Non_Blocking_Read);
   for Ioctl_Cmd_No_Args use
      (Eth_Blocking_Read     => 16#0501#,
       Eth_Non_Blocking_Read => 16#0502#);

   type Ioctl_Cmd_Addr is
     (Eth_Hardware_Address);
   for Ioctl_Cmd_Addr use
      (Eth_Hardware_Address => 16#0500# );

   type Ioctl_Cmd_Protocol is
     (Set_Protocol_Filter,
      Get_Protocol_Filter);
   for Ioctl_Cmd_Protocol use
      (Set_Protocol_Filter => 16#0503#,
       Get_Protocol_Filter => 16#0504#);

   --  ------------------------
   --  Ethernet 'Ioctl' Data
   --  ------------------------

   type Eth_Addr_Ioctl_Arg is array (1..6) of Basic_Integer_Types.Unsigned_8;

   subtype Eth_Proto_Arg is Basic_Integer_Types.Unsigned_16;

   type Eth_Ioctl_Null is private;


end Ethernet_Driver;
