/*----------------------------------------------------------------------------
 *-------------------------      M a R T E   O S      ------------------------
 *----------------------------------------------------------------------------
 *                                                             {MARTE_VERSION}
 *
 *                                'n i c . c'
 *
 *                                     C
 *
 *
 * File 'nic.c'                                                       By Chema.
 *                                                          Jose Maria Martinez
 * Ethernet basic driver functions 
 *
 * {MARTE_COPYRIGHT}
 *
 *---------------------------------------------------------------------------*/
/* {<MAR}
 * Chema:
 * Date format DD-MM-YY:
 *
 *
 * 16-12-02: Total re-make to be similar to the nic.h on Etherboot
 *
 * 26-11-02: First draft.
 *
 *
 * {MAR>} -------------------------------------------------------------------*/
/*****************************************************************************/
/* This module is based on nic.c from the Etherboot proyect:                 */
/* http://etherboot.sourceforge.net v.5.1.3                                  */


/**************************************************************************
Etherboot -  BOOTP/TFTP Bootstrap Program

Author: Martin Renters
  Date: Dec/93
**************************************************************************/

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */


#include "eth_defs.h"
#include "nic.h"
#include "rx_queue.h"
#include "dev.h"
#include "ethernet.h"
#include "pio.h"
#include <drivers/eth_ioctl.h>
//jl #include <misc/hwinterrupts.h> 
#include <intr.h> //jl
#include <net/if_ether.h>


#define ETH_POLL_ERR     -1

struct arptable_t	arptable[MAX_ARP];

/* Currently no other module uses rom, but it is available */
struct rom_info		rom;
/* static unsigned long	netmask; */


/* We define here the reception global queues. It is probably not the */
/* proper place to put them but it's not so bad. */
/* rx_ring_queue give us one rx ring queue for each ethernet interface.*/

static ring_queue_t rx_ring_queue;


/*Here we store the number of pending rx paquets. */

volatile int pending_rx_paquets=0;

/* Only one open and one close per process. */
boolean device_opened=FALSE;

static int dummy(void *a)
{
	return (0);
}

static void dummy_handler(void){

  printf("Handler function not installed\n");

}

static char	packet[ETH_FRAME_LEN];

struct nic	nic =
{
	{
		0,				/* dev.disable */
		{
			DEV_ID_SIZE-1,
			0,
			5,
			PCI_BUS_TYPE,
			0,
			0
		},				/* dev.devid */
		0,				/* index */
		0,				/* type */
		PROBE_FIRST,			/* how_pobe */
		PROBE_NONE,			/* to_probe */
		0,				/* failsafe */
		{},				/* state */
	},
	(int (*)(struct nic *))dummy,		/* poll */
	(void (*)(struct nic *, const char *,
		unsigned int, unsigned int,
		const char *))dummy,		/* transmit */
	(void (*))dummy_handler,                /* interrupt handler*/
	(int (*)(struct nic *))dummy,		/* open */
	0,					/* flags */
	&rom,					/* rom_info */
	//	{00,00,00,00,00,00},	        	/* node_addr */
	arptable[ARP_CLIENT].node,  
	packet,					/* packet */
	0,					/* packetlen */
	0,					/* priv_data */
};



/* eth_poll : Polls the ethernet interface to get a rx packet               */
/*            and stores it in 'data'. 'data' might point a free reserved   */
/*            space on rx ring queue. I might be used with eth_frame_t type.*/
/*            It also stores in len the data length of data                 */
/*            Returns the number of receibed bytes.                         */
/*            On error returns ETH_POLL_ERR when this come up the returned  */
/*            values from the driver are stored in *data and *len           */
/*            Both *data and *len supposed to point to free reserved area   */
/*            inside the reception (ring) buffer.                           */
/*            This function is supposed to be used inside the interrupt     */
/*            handler.                                                      */


/* int eth_poll(char *data, unsigned int *len) */
int eth_poll(char *data)
{
  int ret;
  /* With this, the driver writes directly into the buffer*/
  nic.packet=data;
  ret=(*nic.poll)(&nic);
  //*len=nic.packetlen;
  //return *len;
  if (ret)
    return nic.packetlen;
  return ret;
}


/* eth_transmit : Transmits a packet and waits for completion or timeout.   */
/*                Arguments:                                                */
/*                           dest_addr:    The destination ethernet address.*/
/*                           eth_protocol: Ethernet protocol type.          */
/*                           data_len:     The size of the Data             */
/*                           data:         The data for the ethernet packet.*/

void eth_transmit( const char *dest_addr, 
		   unsigned int eth_protocol, 
		   unsigned int data_len, const char *data)
{
	(*nic.transmit)(&nic, dest_addr, eth_protocol, data_len, data);
}


int eth_handler(void *area, intr_t intr){
  
  eth_frame_t *E;
  eth_frame_t temp;

  if( (*nic.handler)() ){

    if( (E=alloc_ring_element(&rx_ring_queue)) == RING_FULL ){
      printf("Reception Buffer Overflow. Consider Redimension it."
	     "Packet LOST\n");
      temp.info_length=eth_poll(temp.eth_frame);
      return POSIX_INTR_HANDLED_NOTIFY;
    }
    E->info_length=eth_poll(E->eth_frame);
    sem_post(&sem);
    return POSIX_INTR_HANDLED_NOTIFY;
  }
  return POSIX_INTR_HANDLED_NOTIFY;
}//end function eth_handler


ssize_t eth_create(int create_arg)
{
   
  //nic.dev.how_probe = PROBE_FIRST;
  //nic.dev.type = NIC_DRIVER;

  if(sis900_probe(&nic.dev)){
    //    if(hwinterrupts_install_handler (nic.irq,eth_handler) )
    if(posix_intr_associate(nic.irq, eth_handler, NULL, 0) )
      printf("eth_probe: Error installing IRQ handler %d\n",nic.irq);
    else
      printf("eth_probe: IRQ %d\n",nic.irq);


  /* We inicialice the Rx queue. */
  init_ring(&rx_ring_queue);
  /* We inicialize ticks_per_sec */
  ticks_per_sec = ticks_per_second();
  tx_timeout = ticks_per_sec;
  return 0;

  }
  return -1;

}

/******
  Open
*******/

ssize_t eth_open(int file_descriptor, int file_access_mode){
  
  if(!device_opened){
    (*nic.open)(&nic);  
    posix_intr_unlock (nic.irq);   
    //hwinterrupts_enable(nic.irq);
    device_opened=TRUE;
  }
  return 0;
}

/******
  Read
*******/

ssize_t eth_read(int file_descriptor, void *msg, size_t len){
  
  //  struct ethhdr *eth_header;
  //  char *eth_data;
  unsigned int data_len;

  if( (data_len=read_ring(&rx_ring_queue,msg)) < 0 )
    return -1;
  else     
    return data_len;
}//end function eth_read

/*******
  Write
********/

ssize_t eth_write(int file_descriptor, void *msg, size_t len){
  
   struct ethhdr eth_header;
   char *eth_data;
   int i;
   i=0;
   eth_header = *(struct ethhdr *)msg;
   eth_data = (char *) (msg + sizeof(struct ethhdr));
   
/*    eth_transmit(dest_addr,protocol,data_len,data); */
   eth_transmit(eth_header.h_dest,   // destination address. 
		eth_header.h_proto,  // Higher level protocol.
		len - ETH_HLEN,       // data size.
		eth_data);            // data.
  return len;

}

/*******
  Ioctl
********/

ssize_t eth_ioctl(int file_descriptor, ssize_t request, void *priv){

  switch(request){

  case ETH_HWADDR:
    memcpy(priv,nic.node_addr,ETH_ALEN);
    break;
  case ETH_BLOCKING_READ:
    rx_ring_queue.blocking_read=TRUE;
    break;
  case ETH_NON_BLOCKING_READ:
    rx_ring_queue.blocking_read=FALSE;
    break;
  case SET_PROTOCOL_FILTER:
    cli();
    protocol_filter=*(unsigned short *)priv;
    sti();
    break;
  case GET_PROTOCOL_FILTER:
    *(unsigned short *)priv=protocol_filter;
    break;

  default:
    break;
  }
  return 0;
}


void disable(struct dev *dev)
{
	if (dev->disable) {
		dev->disable(dev);
		dev->disable = 0;
	}
}


/********
  Close
*********/
int  eth_close(int file_descriptor)
{
  if(device_opened){
    disable(&nic.dev);
    device_opened=FALSE;
  }
  return 0;
}
