/*----------------------------------------------------------------------------
 *-------------------------      M a R T E   O S      ------------------------
 *----------------------------------------------------------------------------
 *                                                             {MARTE_VERSION}
 *
 *                            'r x _ q u e u e . c'
 *
 *                                     C
 *
 *                                               
 * File 'rx_queue.c'                                                  By Chema.
 *                                                          Jose Maria Martinez
 *                                                         <martinjm@unican.es>
 * Body fuctions of the ring queue (see description). 
 *
 * {MARTE_COPYRIGHT}
 *
 *---------------------------------------------------------------------------*/
/* {<MAR}
 * Chema:
 * Date format DD-MM-YY:
 *
 *
 * 22-11-02: First draft. (lacks of a race condition protection)
 *
 *
 * {MAR>} -------------------------------------------------------------------*/

/* This module implements a ring buffer. Concurrent access is avoid if     */
/* exactly one producer and exactly one consumer are accesing. The reading */
/* process is waiting to consume data that is produced at interrupt time.  */
/* An important consideration of the ring buffer is that we consider the   */
/* buffer empty when the head (current_pos) is == tail (free_pos), and full*/
/* when the (tail + 1) % RING_BUFFER_LEN == head (we waste one element)    */

#include "rx_queue.h"
#include "pio.h"
#include <unistd.h>
#include <time.h>

//static pthread_cond_t the_condition;
//static pthread_mutex_t the_mutex;

void init_ring(ring_queue_t *F){

  //  pthread_mutexattr_t attr;

  F->free_pos=0;
  F->current_pos=0;
  /*We assume that gcc initialices the array to 0s*/
  F->blocking_read=TRUE;
  // We inicialize the semaphore.
  sem_init(&sem, 0,0);

}//end function init_ring.


boolean is_ring_empty(ring_queue_t *F){

  return (F->free_pos==F->current_pos);

}

int insert_ring(ring_queue_t *F, eth_frame_t *E){

  if((F->free_pos +1) % RING_BUFFER_LEN == F->current_pos){
/*     return RING_FULL; */
    return -1;
  }else {
    memcpy(F->ring_buffer[F->free_pos].eth_frame,E->eth_frame,E->info_length);
    F->ring_buffer[F->free_pos].info_length=E->info_length;
    F->free_pos=(F->free_pos +1) % RING_BUFFER_LEN;
    return 0;
  } //end if-else clause


}

eth_frame_t *alloc_ring_element(ring_queue_t *F){

  eth_frame_t *ret;

  if((F->free_pos +1) % RING_BUFFER_LEN == F->current_pos){
    return RING_FULL;
  }else {
    ret = (eth_frame_t *) &F->ring_buffer[F->free_pos];
    F->free_pos=(F->free_pos +1) % RING_BUFFER_LEN;
    return ret;
  } //end if-else clause

}// end function alloc_ring_element



int read_ring(ring_queue_t *F, char *E ){
  
  int ret;
  // the sleep a msleep (1 ms).
  //  struct timespec sleep_time_us = { 0, 100e3};

  /* FALTA : inhibir interrupciones.*/
  //  cli();
  while(is_ring_empty(F)){
    if(F->blocking_read==FALSE){
      /* FALTA DESHINIBIR LAS INTERRUPCIONES*/
      //sti();
      return RING_EMPTY;
    }

    /* FALTA IMPLEMENTAR EL WAIT ESPECIAL QUE INHIBE INT.*/
    /* SUSTITUIR EL SLEEP POR EL WAIT*/
    //sti();
    //wait();
    //nanosleep(&sleep_time_us,NULL);
    //cli();
    sem_wait(&sem);
  }
  ret=F->ring_buffer[F->current_pos].info_length;
  memcpy(E, F->ring_buffer[F->current_pos].eth_frame, ret);
  F->current_pos = (F->current_pos +1) % RING_BUFFER_LEN;

  /* FALTA DESIHINIBIT LAS INTERRUPCIONES.*/
  //sti();
  return ret;

}//end function read_ring
