/*----------------------------------------------------------------------------
 *-------------------------      M a R T E   O S      ------------------------
 *----------------------------------------------------------------------------
 *                                                             {MARTE_VERSION}
 *
 *                            'r x _ q u e u e . h'
 *
 *                                     C
 *
 *                                               
 * File 'rx_queue.h'                                                  By Chema.
 *                                                          Jose Maria Martinez
 *                                                         <martinjm@unican.es>
 * Body fuctions on 'rx_queue.c'
 *
 * {MARTE_COPYRIGHT}
 *
 *---------------------------------------------------------------------------*/
/* {<MAR}
 * Chema:
 * Date format DD-MM-YY:
 *
 *
 * 22-11-02: First draft. 
 *
 *
 * {MAR>} -------------------------------------------------------------------*/

#ifndef _MARTE_RX_QUEUE_H
#define _MARTE_RX_QUEUE_H


/* This module implements a RING ring queue. It is intended */
/* to be a RING reception queue with a ring storage behavior*/

#include "eth_defs.h"
#include <semaphore.h>

#define RING_EMPTY -1
#define RING_FULL  NULL

typedef struct {
  unsigned char eth_frame[ETH_FRAME_LEN];
  unsigned int info_length; 
// "unsigned int" is impossed by packetlen in nic.h
}eth_frame_t; 

typedef struct {
  volatile unsigned short int free_pos; /*The position to be filled.*/
  unsigned short int current_pos; /*The first position to be read.*/
  eth_frame_t ring_buffer[RING_BUFFER_LEN];
  /* blocking_read is used to decide if the read operation is blocking*/
  /* de default is blocking. */
  boolean blocking_read;
} ring_queue_t;

sem_t sem;

/*init_ring : This function inicialices the RING queue.*/
/*            you MUST call it BEFORE any access to the queue*/
extern void init_ring(ring_queue_t *F);

/*is_ring_empty : Checks if the queue is empty. Returns FALSE if*/
/*                thereis any frame to be read and TRUE if the  */
/*                queue is empty. */
extern boolean is_ring_empty(ring_queue_t *F);



/* inset_ring : Stores E in the ring buffer F. The free_pos & frames_num */
/*              variables are updated.                                   */ 
/*              Return 0 if OK.                                          */
/*              RING_FULL if no room in buffer.                          */

extern int insert_ring(ring_queue_t *F, eth_frame_t *E);


/* alloc_ring_element : Will return a free eth_frame_t possition on the  */
/*                      ring buffer to write a frame. It is used to loss */
/*                      one memory copy.                                 */

extern eth_frame_t *alloc_ring_element(ring_queue_t *F );


/*read_ring : Copies in E the the first-in-frame in the buffer.*/
/*            Returns the number of rx bytes. Updates current_pos */
/*            variable.It works under interrupts inhibited. */
/*            this function is blocking y default. If it is configured*/
/*            non blocking QUEUE_EMPTY is returned when empty.*/
extern int read_ring(ring_queue_t *F, char *E );



#endif /*_MARTE_RX_QUEUE_H*/
