------------------------------------------------------------------------------
--                                                                          --
--            FLORIST (FSU Implementation of POSIX.5) COMPONENTS            --
--                                                                          --
--           P O S I X . S H A R E D _ M E M O R Y _ O B J E C T S          --
--                                                                          --
--                                  B o d y                                 --
--                                                                          --
--                                                                          --
--  Copyright (c) 1996, 1997            Florida  State  University  (FSU),  --
--  All Rights Reserved.                                                    --
--                                                                          --
--  This file is a component of FLORIST, an  implementation of an  Ada API  --
--  for the POSIX OS services, for use with  the  GNAT  Ada  compiler  and  --
--  the FSU Gnu Ada Runtime Library (GNARL).   The  interface  is intended  --
--  to be close to that specified in  IEEE STD  1003.5: 1990  and IEEE STD  --
--  1003.5b: 1996.                                                          --
--                                                                          --
--  FLORIST is free software;  you can  redistribute  it and/or  modify it  --
--  under terms of the  GNU  General  Public  License as  published by the  --
--  Free Software Foundation;  either version  2, or (at  your option) any  --
--  later version.  FLORIST is distributed  in  the hope  that  it will be  --
--  useful, but WITHOUT ANY WARRANTY;  without  even the implied  warranty  --
--  of MERCHANTABILITY or FITNESS FOR A PARTICULAR  PURPOSE.  See  the GNU  --
--  General Public License for more details.  You  should have  received a  --
--  copy of the GNU General Public License  distributed  with  GNARL;  see  --
--  file  COPYING.  If not,  write to  the  Free  Software  Foundation, 59  --
--  Temple Place - Suite 330, Boston, MA 02111-1307, USA.                   --
--                                                                          --
--  As a special exception, if other files instantiate generics from  this  --
--  unit, or you link this unit with other files to produce an  executable, --
--  this  unit does not by itself cause the  resulting  executable  to  be  --
--  covered  by the  GNU  General  Public License. This exception does not  --
--  however invalidate any other  reasons why the executable file might be  --
--  covered by the GNU Public License.                                      --
--                                                                          --
------------------------------------------------------------------------------
--  [$Revision: 1.1.1.1 $]

with Ada.Text_IO; use Ada.Text_IO;
--  with POSIX.C;
with POSIX.IO;
with POSIX.Implementation;
--  with POSIX.Permissions.Implementation;
with Unchecked_Conversion;

-- jl ----------------------------
--  with Basic_Integer_Types;
--  use Basic_Integer_Types;
with Interfaces.C;
use Interfaces.C;
-- jl ----------------------------


package body POSIX.Shared_Memory_Objects is

--   use POSIX.C;
   use POSIX.Implementation;
--    use POSIX.Permissions.Implementation;

--  jl------------------------------------------
--  jl de /kernel/option_sets.ads
   type Bits is mod 2**32;
   for Bits'Size use Int'Size;

   type mode_t is mod 2**32;
   for mode_t'Size use 32;

   subtype Char is POSIX_Character;
   type Char_Ptr is access constant Char;
   pragma Convention (C, Char_Ptr);

   O_RDONLY : Bits := 1;
   O_WRONLY : Bits := 2;
   O_RDWR   : Bits := 3;
   O_CREAT  : Bits := 4;

   MI_PERMISO : mode_t := 3;
   Mi_Oflag : Integer := 9;

   function shm_unlink (name : char_ptr) return int;
   pragma Import (C, shm_unlink, "shm_unlink");

   function shm_open
     (name  : char_ptr;
      oflag : int;
      mode  : mode_t) return int;
   pragma Import (C, shm_open, "shm_open");

--   function To_int is new Unchecked_Conversion (Bits, int);
   function To_int is new Unchecked_Conversion (Bits, Int);

   C_File_Mode : constant array (POSIX.IO.File_Mode) of Bits :=
     (POSIX.IO.Read_Only  => O_RDONLY,
      POSIX.IO.Write_Only => O_WRONLY,
      POSIX.IO.Read_Write => O_RDWR);

   --------------------------
   --  Open_Shared_Memory  --
   --------------------------

   function Open_Shared_Memory
     (Name           : POSIX.POSIX_String;
      Mode           : POSIX.IO.File_Mode;
      Options        : POSIX.IO.Open_Option_Set := --  POSIX.IO.Empty_Set;
         POSIX.IO.Open_Option_Set (POSIX.IO.Empty_Set);
      Masked_Signals : POSIX.Signal_Masking := POSIX.RTS_Signals)
     return POSIX.IO.File_Descriptor is

      Name_With_NUL : POSIX_String (1 .. Name'Length + 1);
      Result : POSIX.IO.File_Descriptor;

      Mi_Resultado : Integer := 0;

   begin
      Name_With_NUL (1 .. Name'Length) := Name (1 .. Name'Length);
      Name_With_NUL (Name'Length + 1) := POSIX_Character (Nul);


      Mi_Resultado := Shm_Open (
            Name_With_NUL (Name_With_NUL'First)'Unchecked_Access,  -- name
            Mi_Oflag,  --  oflag
            Mi_Permiso);

      Put_Line (Mi_Resultado'Img);
      Result := POSIX.IO.File_Descriptor (Mi_Resultado);


--        Check (
--          Shm_Open (
--            Name_With_NUL (Name_With_NUL'First)'Unchecked_Access,  -- name
--            Mi_Oflag,  --  oflag
--            Mi_Permiso
--          ) -- shm_open
--        ) -- check
--      );   -- POSIX.IO...
      return Result;
   end Open_Shared_Memory;

   ------------------------------------
   --  Open_Or_Create_Shared_Memory  --
   ------------------------------------

   function Open_Or_Create_Shared_Memory
     (Name           : POSIX.POSIX_String;
      Mode           : POSIX.IO.File_Mode;
      Permissions    : POSIX.Permissions.Permission_Set;
      Options        : POSIX.IO.Open_Option_Set := --  POSIX.IO.Empty_Set;
         POSIX.IO.Open_Option_Set (POSIX.IO.Empty_Set);
      Masked_Signals : POSIX.Signal_Masking := POSIX.RTS_Signals)
      return POSIX.IO.File_Descriptor is

      Name_With_NUL : POSIX_String (1 .. Name'Length + 1);
      Result : POSIX.IO.File_Descriptor;

   begin
      Name_With_NUL (1 .. Name'Length) := Name (1 .. Name'Length);
      Name_With_NUL (Name'Length + 1) := POSIX_Character (Nul);

      Result := POSIX.IO.File_Descriptor (
        Check (
          Shm_Open (
            Name_With_NUL (Name_With_NUL'First)'Unchecked_Access,  -- name
            Mi_Oflag,  --  oflag
            Mi_Permiso
          ) -- shm_open
        ) -- check
      );   -- POSIX.IO...
      return Result;
   end Open_Or_Create_Shared_Memory;

   ----------------------------
   --  Unlink_Shared_Memory  --
   ----------------------------

   procedure Unlink_Shared_Memory
     (Name : in POSIX.POSIX_String) is
      Name_With_NUL : POSIX_String (1 .. Name'Length + 1);

   begin
      Name_With_NUL (1 .. Name'Length) := Name (1 .. Name'Length);
      Name_With_NUL (Name'Length + 1) := POSIX_Character (Nul);

      Check (
        shm_unlink
         (
          Name_With_NUL (Name_With_NUL'First)'Unchecked_Access
         ) -- shm_unlink
      ); -- check
     end Unlink_Shared_Memory;

end POSIX.Shared_Memory_Objects;
