/*----------------------------------------------------------------------------
 *-------------------------         R T - E P         ------------------------
 *----------------------------------------------------------------------------
 *                                                             {MARTE_VERSION}
 *
 *                          'l n k _ c o n f i g . c'
 *
 *                                     C
 *
 *                                               
 * File 'lnk_config.c'                                                By Chema.
 *                                                          Jose Maria Martinez
 *                                                         <martinjm@unican.es>
 *
 * Body functions of lnk_config.h configuration functions.
 *
 *
 *---------------------------------------------------------------------------*/
 /*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */
/*****************************************************************************/
/* Object implementation                                                     */
/*****************************************************************************/

 
#include <string.h>
#include "prot_types.h"
#include "lnk_config.h"
#include "ring_config.h"

/* elements variable inside st_array will  store the number of        */
/* stations in the array. Be careful! the array begins in 0 to        */
/* elements -1                                                        */
typedef struct {
  lnk_cell_t stations_array[MAX_STATIONS];
  positive_t elements; 
} st_array;

/* st_config will store the ring configuration                        */
static st_array st_config;

int init_config(){
	
  int i, number_of_stations;

  number_of_stations=( sizeof(stations)/sizeof(stations[0]) );
  /* We inicialize the stations counter.*/
  st_config.elements=0;
  /* If we have more stations than the MAX_STATIONS limit...        */
  if( number_of_stations >= MAX_STATIONS )
    return STATIONS_OVERFLOW;
 

  for(i=0;i<number_of_stations;i++){

#ifdef LNK_CONFIG_DEBUG_MODE
    printf("lnk_config:init_config> Have read: %s from configuration\n",stations[i]);
#endif

    *(struct ether_addr *)st_config.stations_array[st_config.elements].addr_dest=
      *(struct ether_addr *)ether_aton(stations[i]);

    st_config.stations_array[st_config.elements].valid_station=TRUE;
#ifdef LNK_CONFIG_DEBUG_MODE

    printf("lnk_config:init_config> Converted to: %s\n",
	   ether_ntoa((struct ether_addr *)
		      st_config.stations_array[st_config.elements].addr_dest));
#endif
  
    st_config.elements++;
  } //end for.

  return 0;
}//end function init

int read_station(lnk_cell_t *sta, positive_t node){

  if(st_config.elements > node){
    *sta=st_config.stations_array[node];
    if(sta->valid_station)
      return 0;
    else
      return STA_NOT_VALID;
  }else
    return STA_NOT_FOUND;

} //end read_station


positive_t number_nodes(){

  return st_config.elements;

}//end number_nodes


int invalidated_station(unsigned char *st_addr){

  int i;
  for (i=0;i<st_config.elements;i++){
    if(st_config.stations_array[i].valid_station==FALSE){
      if(addrcmp(st_config.stations_array[i].addr_dest,st_addr,ETH_ALEN))
	return STA_NOT_VALID;
    } // end if.

  }//end for.
  return 0;
} //end invalidated_station fuction.



int check_station(unsigned char *st_addr){

  lnk_cell_t sta;
  int i;
  for (i=0;i<st_config.elements;i++){
    sta=st_config.stations_array[i];
    if(addrcmp(sta.addr_dest,st_addr,ETH_ALEN)){
      if(sta.valid_station)
	return 0;
      else
	return STA_NOT_VALID;
    }//end if
  }//end for.
  return STA_NOT_FOUND;
}//end check_station

int inhibit_station(unsigned char *st_addr){
  
  lnk_cell_t sta;
  int i;
  for (i=0;i<st_config.elements;i++){
    sta=st_config.stations_array[i];
    if(addrcmp(sta.addr_dest,st_addr,ETH_ALEN)){
      st_config.stations_array[i].valid_station=FALSE;
      return 0;
    }//end if
  }//end for.
  return STA_NOT_FOUND;

} //end inhibit_station;



