/*----------------------------------------------------------------------------
 *-------------------------         R T - E P         ------------------------
 *----------------------------------------------------------------------------
 *                                                             {MARTE_VERSION}
 *
 *                          'l n k _ c o n f i g . h'
 *
 *                                     C
 *
 *                                               
 * File 'lnk_config.h'                                                By Chema.
 *                                                          Jose Maria Martinez
 *                                                         <martinjm@unican.es>
 *
 * Ring configuration functions.
 *
 *
 *---------------------------------------------------------------------------*/
 /*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */
/**********************************************************************/
/*         lnk_config -> Configuration Object.                        */
/*         FILE_NAME_CONFIG might be a text file with                 */
/*         the followings fields. (see prot_types.h in order to set   */
/*                               FILE_NAME_CONFIG)                    */
/*         It is assumed the interface is in eth0                     */
/*         for any changes on the interface look into prot_types.h    */
/*                                                                    */
/*   File structure:                                                  */
/*                                                                    */
/*   mac_address1                                                     */
/*   mac_address2                                                     */
/*  i.e...                                                            */
/*   00:4F:4C:02:F6:68                                                */
/*   00:4F:49:04:7D:8D                                                */
/**********************************************************************/

#ifndef FILE_LNK_CONFIG_H
#define FILE_LNK_CONFIG_H

#include <stdio.h>
#include <net/if_ether.h>
#include "prot_types.h"

/* We need to strore the network configuration in order to be         */
/* accesible to other functions.                                      */ 

typedef struct node {
  unsigned char addr_dest[ETH_ALEN];
  boolean valid_station;
} lnk_cell_t;


extern int init_config();
/* init_config will inicialize the whole structure of node            */
/* information, it returns:                                           */ 
/*  - 0 if OK.                                                        */
/*  - STATIONS_OVERFLOW if there are more stations than the amount we */
/*        can handle                                                  */
/*  - FILE_ERROR if there was a problem opening the configuration.    */


extern int read_station(lnk_cell_t *sta, positive_t node);
/* read_station returns the station sta placed on node position on the */
/* st_array configuration returns:                                     */
/*  - 0 if OK                                                          */
/*  - STA_NOT_FOUND if the station isn't in the configuration structure*/
/*  - STA_NOT_VALID if the station is inhibited but stores the         */
/*       information about the station in sta                          */ 
/* The first station is node 0                                         */

extern positive_t number_nodes();
/* number_nodes will return the numbers of nodes in the logical ring.  */ 

extern int check_station(unsigned char *st_addr);
/* check_station checks the st_addr station if it is active, returns:  */
/*  - 0 if it is a valid station.                                      */
/*  - STA_NOT_VALID if the station is set not valid.                   */
/*  - STA_NOT_FOUND if the station isn't on the logical ring           */

extern int invalidated_station(unsigned char *st_addr);
/* invalidated_station will check on the configuration if st_addr is a */
/* invalid station. BE CAREFUL using this function, it doesn't check   */
/* if the station isn't on the ring or if it is valid, it only check   */
/* if one of the invalid marked station is st_addr.                    */
/* This is the fastest solution to check if st_addr is an invalid      */
/* station.                                                            */
/* Returns:                                                            */
/*  - 0 if it isn't on the invalidated stations.                       */
/*  - STA_NOT_VALID if the station IS amoung th invalidated stations   */

extern int inhibit_station(unsigned char *st_addr);
/* inhibit_station will inhibit a station from the ring, returns:      */
/*  - 0 if OK                                                          */
/*  - STA_NOT_FOUND if the station isn't on the ring                   */
/* inhibiting a previous inhibited station isn't an error.             */
#endif /*FILE_LNK_CONFIG_H*/

