/*----------------------------------------------------------------------------
 *-------------------------         R T - E P         ------------------------
 *----------------------------------------------------------------------------
 *                                                             {MARTE_VERSION}
 *
 *                            'm a r t e l n k . c'
 *
 *                                     C
 *
 *                                               
 * File 'martelnk.c'                                                  By Chema.
 *                                                          Jose Maria Martinez
 *                                                         <martinjm@unican.es>
 *
 * Main communication thread (Body function of martelnk.h).
 *
 *
 *---------------------------------------------------------------------------*/
 /*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <time.h>
#include <errno.h>
#include <net/osdep.h>
#include <pthread.h>
#include <drivers/eth_ioctl.h>
#include <fcntl.h>

//Developer defined includes
#include "prot_types.h"
#include "prio_queue_monitor.h"
#include "lnk_config.h"
#include "martelnk.h"


/* SHORT_INT is the max. range of a a short int                       */
#define SHORT_INT 65536


/* we have to set this global structure to be used in the             */
/* function below, since its a way to know the current                */
/* station address.                                                   */

/* struct ether_addr                                                  */
/* {                                                                  */
/*   unsigned char	ether_addr_octet[ETH_ALEN];                   */
/* };                                                                 */

static struct ether_addr current_sta_addr;

static struct ether_addr addr_suc_station;

/* We also have to store the station possition on the ring for future */
/* modifications on the ring due to station failure. positive_t puts  */
/* a maximum of 32k stations on the ring.                             */

static positive_t station_possition; 

// This is for the ring_handler function.

int file_descriptor_error;

/* We have to make this enum type global since it is going            */
/* to be used by ring_handler()                                       */
/* State enumeration                                                  */
static enum states {offline_state,	 
	     idle_state,
	     send_read_token_state,
	     check_token_state,
	     send_permission_state,
	     send_token_state,
	     send_info_state,
	     recv_info_state
} last_station_state=offline_state; 


/* info_packet will store the information to transmit and token_packet is   */
/* the token                                                          */
static info_packet_t info_packet;
static token_packet_t token_packet;

static char tx_buffer[ETH_FRAME_LEN];
static unsigned short int last_packet_number=0;


/* In this global variable we will store the retries done when the    */
/* timeout pops up.:)                                                 */
static int retry_counter=0;

/* This boolean will tell the station if it has to check the media in */
/* case of error.                                                     */
static boolean check_for_error=FALSE;
/* token_master will set the pass-token-management-station.           */
static boolean token_master=FALSE;

/* check_for_valid_station will indicate the station that has found   */
/* the failure station to reassign the possitive valid_station to     */
/* zero when the token get one round                                  */

static boolean check_for_valid_station=FALSE;

/* Global signal variables to manage the timer                        */
static   struct itimerspec timerdata;
static   timer_t timer_id;

/* time_out enable and disable.                                       */

static struct timespec timer_disable;
static struct timespec timeout;



int read_token(token_packet_t *my_token, queue_t *Q){

/* read_token reads the hightest priority element in the queue Q      */
/* and write in *my_token the token we will tx to the next station.   */
/* If we read no element from the queue-> QUEUE_EMPTY a generic       */
/* holding-link token will be created                                 */ 
/* The function returns 0 if OK.                                      */

  element_t *E;
  priority_t P;
  int *ret_val;

#ifdef MARTELNK_DEBUG_MODE
  printf("martelnk:read_token> Trying to make a token from the queue...");
#endif

  switch(read_first_in_queue(&E,&P,Q,TX_QUEUE)){
  case QUEUE_EMPTY:
    my_token->packet_id='T';
    my_token->priority=0;

    *(struct ether_addr * )my_token->addr_prior = current_sta_addr;

    *(struct ether_addr * )my_token->eth_header.h_dest = addr_suc_station;

    *(struct ether_addr * )my_token->eth_header.h_source = current_sta_addr;

    my_token->eth_header.h_proto=htons(RT_EP_PROTOCOL);

    break;
  case SYNCH_ERROR:
    printf("martelnk:read_token> Synch error in monitor\n");
    *ret_val=SYNCH_ERROR;
    pthread_exit(ret_val);
    break;
  case 0:
    my_token->packet_id='T';
    my_token->priority=P;

    *(struct ether_addr * )my_token->addr_prior = current_sta_addr;

    *(struct ether_addr * )my_token->eth_header.h_dest = addr_suc_station;

    *(struct ether_addr * )my_token->eth_header.h_source = current_sta_addr;

    my_token->eth_header.h_proto=htons(RT_EP_PROTOCOL);
    break;
  default:  
    printf("martelnk:read_token> Unexpected error\n");
    *ret_val=UNEXPECTED_ERROR;
    pthread_exit(ret_val);
    break;
  }//end switch


  /* As the station becomes the token_master station, it stores its   */ 
  /* address on the token                                             */

  *(struct ether_addr *)my_token->token_master_addr = current_sta_addr;

  /* We asign the data sequence number.                               */
  my_token->packet_number=(last_packet_number+1) % SHORT_INT;

#ifdef MARTELNK_DEBUG_MODE
 printf("OK!\n");
 printf("martelnk:read_token> Have Read: %d priority\n", P);
#endif
 return 0;
}//end - function read_token.

void read_info( info_packet_t *my_info, queue_t *Q, element_t **E){

/* read_info reads from the Tx priority queue the element to tx,      */
/* writes in my_info the header of th tx info packet and it returns   */
/* also the element to tx.                                            */
  priority_t P;
  int *ret_val;
#ifdef MARTELNK_DEBUG_MODE
  printf("martelnk:read_info>Trying to read the info of the queue...");
#endif

  switch(queue_extract(E,&P,Q,TX_QUEUE)){
  case SYNCH_ERROR:
    printf("martelnk:read_info>Synch error in monitor\n");
    *ret_val=SYNCH_ERROR;
    pthread_exit(ret_val);
    break;
  case 0:
    my_info->packet_id = 'I';
    my_info->priority = P;
    my_info->channel_id = ((element_t *)*E)->channel_id;
    my_info->length_info = ((element_t *)*E)->element_size;

    *(struct ether_addr *)my_info->eth_header.h_dest=
      *(struct ether_addr *)  ((element_t *)*E)->addr_dest;

    *(struct ether_addr *)my_info->eth_header.h_source = current_sta_addr;

    my_info->eth_header.h_proto=htons(RT_EP_PROTOCOL);
    break;
  default:
    printf("martelnk:read_info>Unexpected error\n");
    *ret_val=UNEXPECTED_ERROR;
    pthread_exit(ret_val);
    break;
  }//end switch

  /* We asign the data sequence number.*/
  my_info->packet_number=(last_packet_number+1) % SHORT_INT;


#ifdef MARTELNK_DEBUG_MODE
 printf("martelnk:read_info> OK!\n");
 printf("martelnk:read_info> INFO: %s\n", 
	((element_t *)*E)->element_pointer);
 printf("martelnk:read_info> STATION: %s\n",
	ether_ntoa((struct ether_addr *)  ((element_t *)*E)->addr_dest));

#endif

}//end - funcion read_info.


void write_info( info_packet_t *his_info, queue_t *Q, char *rx_ptr ){

/* write_info writes the received info package to the queue Q         */ 
/* *rx_ptr is a pointer to the received information.                  */

  element_t *E;
  int *ret_val;

#ifdef MARTELNK_DEBUG_MODE
  printf("martelnk:write_info> Trying to write the info in the Rx queue...");
#endif
  E=alloc_element();
  memcpy(E->element_pointer, rx_ptr + sizeof(info_packet_t),
	 his_info->length_info);
  E->element_size=his_info->length_info;

  /* This address might have benn modified to the sender address in   */
  /* his_info                                                         */

  *(struct ether_addr *)E->addr_dest=
    *(struct ether_addr *) his_info->eth_header.h_source;

  E->channel_id=his_info->channel_id;

  switch(queue_insert(E,his_info->priority,&Q[E->channel_id],E->channel_id)){

  case SYNCH_ERROR:
    printf("martelnk:write_info> Synch error in monitor\n");
    *ret_val=SYNCH_ERROR;
    pthread_exit(ret_val);
    break;

  case 0:
    // All is OK.
    break;

  default:
    printf("martelnk:write_info> Unexpected error\n");
    *ret_val=UNEXPECTED_ERROR;
    pthread_exit(ret_val);
    break;
  }//end switch


#ifdef MARTELNK_DEBUG_MODE
  printf("martelnk:write_info>OK!\n");
  printf("martelnk:write_info>Written %s in the queue.\n",E->element_pointer);
#endif


}//end - funcion write_info.


void write_token( token_packet_t *his_token, queue_t *Q){

  /* write_token takes the token *his_token and check its priority    */
  /* with the priority of the highest element in the queue and return */
  /* the highest priority element of both.                            */

  element_t *E;
  priority_t P;
  int *ret_val;

#ifdef MARTELNK_DEBUG_MODE
  printf("martelnk:write_token> Comparing Rx token with "
	 "the highest priority element in the queue...\n");
#endif

  switch(read_first_in_queue(&E,&P,Q,TX_QUEUE)){
  case QUEUE_EMPTY:
    /* In case that our queue AND the predecessor queue are empty, we */ 
    /* change the addr_prior field in order to circulate the token,   */
    /* in case the predessor hasn't got the queue empty we do nothing */
    /* to the token and we pass it. */
    if(his_token->priority == 0 ){
      *(struct ether_addr *)his_token->addr_prior = current_sta_addr;
    }

      *(struct ether_addr *)his_token->eth_header.h_dest = addr_suc_station;

      *(struct ether_addr *)his_token->eth_header.h_source = current_sta_addr;

      his_token->eth_header.h_proto=htons(RT_EP_PROTOCOL);
    break;
  case SYNCH_ERROR:
    printf("martelnk:write_token> Synch error in monitor\n");
    *ret_val=SYNCH_ERROR;
    pthread_exit(ret_val);
    break;
  case 0:
    if(his_token->priority < P){
      his_token->priority=P;
      *(struct ether_addr *)his_token->addr_prior = current_sta_addr;
    }

      *(struct ether_addr *)his_token->eth_header.h_dest = addr_suc_station;

      *(struct ether_addr *)his_token->eth_header.h_source = current_sta_addr;

      his_token->eth_header.h_proto=htons(RT_EP_PROTOCOL);
    break;
  default:
    printf("martelnk:write_token> Unexpected error\n");
    *ret_val=UNEXPECTED_ERROR;
    pthread_exit(ret_val);
    break;
  }//end switch

  /* We asign the data sequence number.*/
  his_token->packet_number=(last_packet_number+1) % SHORT_INT;

#ifdef MARTELNK_DEBUG_MODE
  printf("OK!\n");
#endif

} //end - funcion write_token.


int net_info(boolean *token_possession){

  /* This fuction reads the configuration of the predecesor, current  */
  /* and succesor station, it also places in the boolean              */
  /* token_possession if the station has the token.                   */
  /* This function returns 0 if OK. */


  lnk_cell_t sta;
  positive_t nodes,i;
  int *ret_val;
  /* first, we inicialize the structures.                              */

  switch(init_config()){

  case FILE_ERROR:
    printf("martelnk:net_info> File error\n");
    *ret_val=FILE_ERROR;
    pthread_exit(ret_val);
    break;
  case STATIONS_OVERFLOW:
    printf("martelnk:net_info> Stations overflow\n");
    *ret_val=STATIONS_OVERFLOW;
    pthread_exit(ret_val);
    break;
  case 0:
    /* All is correct */
    break;
  default:
    printf("martelnk:net_info> Unexpected error\n");
    *ret_val=UNEXPECTED_ERROR;
    pthread_exit(ret_val);
    break;

  }//end switch

  nodes=number_nodes();

  for(i=0;i<nodes;i++){
  	
    if(read_station(&sta,i) < 0)
      exit(UNEXPECTED_ERROR);

    if( addrcmp(sta.addr_dest,current_sta_addr.ether_addr_octet,ETH_ALEN)){
       station_possition=i;
      if(i==0){
	/* we are on the first station of the list. */
	if(read_station(&sta,nodes-1) < 0)
	   exit(UNEXPECTED_ERROR);
	*token_possession=TRUE;
      }else{
	if(read_station(&sta,i-1) < 0 )
	    exit(UNEXPECTED_ERROR);
	*token_possession=FALSE;
      }//end if-else clause.

      if (read_station(&sta,(i+1) % nodes ) <0 )
           exit(UNEXPECTED_ERROR);
      addr_suc_station = *(struct ether_addr *)sta.addr_dest;

      break;
    }//end if.
    
  }//end for.

#ifdef MARTELNK_DEBUG_MODE

      printf("martelnk:martelnk:net_info> Next Station: %s \n",
	     ether_ntoa((struct ether_addr * )&addr_suc_station ));
#endif

      return 0;

}//end function net_info



void ring_handler(){

  int status;
  positive_t nodes;
  lnk_cell_t sta;
  int *ret_val;
  /*signal variables*/
  sigset_t set;
  int received_sig;
#ifdef FAULT_HANDLING_DEBUG_MODE
  int seq_number;
#endif


  while(1){
    sigemptyset(&set);
    sigaddset(&set,SIGRTMAX);
    // un sigwait.!!!
    sigwait(&set,&received_sig);
    /*we have to disable the timer*/
    timerdata.it_interval=timer_disable;
    timerdata.it_value=timer_disable;
  
    if(timer_settime(timer_id,0,&timerdata,NULL)==-1){
      perror("timer_settimer error\n");
      *ret_val=UNEXPECTED_ERROR;
      pthread_exit(ret_val);
    }

#ifdef MARTELNK_DEBUG_MODE
    printf("martelnk:ring_handler>The timer has send the TIMEOUT signal.\n");
    printf("martelnk:ring_handler>Retry_counter: %d - Last Station State:"
	   " %d .\n", retry_counter,last_station_state);
#endif

    
    if(retry_counter < MAX_RETRIES){
 
    /* In case it's the first retry, we suposse the frame is lost or  */
    /* the receiver station takes too long in response. so we         */
    /* retransmit the frame.                                          */
    
    switch(last_station_state){
      
    case send_read_token_state:
    case send_token_state:

#if defined MARTELNK_DEBUG_MODE || defined FAULT_HANDLING_DEBUG_MODE
      printf("martelnk:ring_handler:send_read_token || send_token_state> "
	     "Starting...\n");
#endif
      status=write(file_descriptor_error, &token_packet, sizeof(token_packet));

#if defined MARTELNK_DEBUG_MODE || defined FAULT_HANDLING_DEBUG_MODE
      printf("martelnk:ring_handler:send_read_token || send_token_state> "
	     "Retx Token\n");
#endif
/* status=sendto(socket_srv,&token_packet,sizeof(token_packet),0,NULL,0); */

      if(status < 0) perror("martelnk:send_read_token:>write error!!");
#ifdef FAULT_HANDLING_DEBUG_MODE
      seq_number=token_packet.packet_number;
#endif
      break;
    case send_info_state:

#if defined MARTELNK_DEBUG_MODE || defined FAULT_HANDLING_DEBUG_MODE
      printf("martelnk:ring_handler:send_info_state> Starting recovering...\n");
#endif
      /* We have to check if the failure station is one of the        */
      /* inhibited ones.                                              */
      if(invalidated_station(info_packet.eth_header.h_dest)==STA_NOT_VALID){
#if defined MARTELNK_DEBUG_MODE || defined FAULT_HANDLING_DEBUG_MODE
      printf("martelnk:ring_handler:send_info_state> The station was not a "
	     "valid one.\n \t Transmision of a new token.\n");
#endif
	token_master=TRUE;
	read_token(&token_packet,&Q_tx);
	status =
	  write(file_descriptor_error, &token_packet, sizeof(token_packet));

/*status=sendto(socket_srv,&token_packet,sizeof(token_packet),0,NULL,0); */

	if(status < 0) perror("martelnk:send_read_token:>sendto_function");
	/* This is done to get the final value of 0 in retry_counter   */
	/* to reset the timer.*/ 
	retry_counter=-1; 
	break;
      }
#if defined MARTELNK_DEBUG_MODE || defined FAULT_HANDLING_DEBUG_MODE
      printf("martelnk:ring_handler:send_info_state> Retransmiting the INFO\n");
#endif
      status = write(file_descriptor_error, tx_buffer,
		     sizeof(info_packet) + info_packet.length_info);

/*       status=sendto(socket_srv,tx_buffer, */
/* 		    sizeof(info_packet)+info_packet.length_info,0,NULL,0); */
      if(status < 0) perror("martelnk:send_read_token:>sendto_function");
#ifdef FAULT_HANDLING_DEBUG_MODE
      seq_number=info_packet.packet_number;
#endif
      break;

    case send_permission_state:

      /* We have to check if the failure station is one of the        */
      /* inhibited ones.                                              */
      if(invalidated_station(info_packet.eth_header.h_dest)==STA_NOT_VALID){
#if defined MARTELNK_DEBUG_MODE || defined FAULT_HANDLING_DEBUG_MODE
      printf("martelnk:ring_handler:send_permission_state> The station was "
	     "not a valid one.\n \t Transmision of a new token.\n");
#endif
	token_master=TRUE;
	read_token(&token_packet,&Q_tx);

	status = 
	  write(file_descriptor_error, &token_packet, sizeof(token_packet) );
/* 	status=sendto(socket_srv,&token_packet,sizeof(token_packet),0,NULL,0); */

	if(status < 0) perror("martelnk:send_read_token:>sendto_function");
	/* This is done to get the final value of 0 in retry_counter   */
	/* to reset the timer.*/ 
	retry_counter=-1; 
	break;
      }
#if defined MARTELNK_DEBUG_MODE || defined FAULT_HANDLING_DEBUG_MODE
      printf("martelnk:ring_handler:send_permission_state> "
	     "Retransmiting the Transmit Token\n");
#endif
      status = 
	write(file_descriptor_error, &token_packet, sizeof(token_packet) );
/*  status=sendto(socket_srv,&token_packet,sizeof(token_packet),0,NULL,0); */
      if(status < 0) perror("martelnk:send_read_token:>sendto_function");
#ifdef FAULT_HANDLING_DEBUG_MODE
      seq_number=token_packet.packet_number;
#endif
      break;
    default:
      //This case is never reached.
      break;
    }//end switch(last_station_state)

    /* Since there has been retry, */
    retry_counter++;
    
  }else{
  
    /* In case it's the last retry, we bypass the successor and makes */
    /* the new successor the successor of the successor.              */


    switch(last_station_state){

    case send_read_token_state:
    case send_token_state:

      check_for_valid_station=TRUE;
      token_packet.failure_station=1;
      *(struct ether_addr *) token_packet.failure_station_addr = 
	addr_suc_station;

      nodes=number_nodes();
      inhibit_station(addr_suc_station.ether_addr_octet);
#if defined MARTELNK_DEBUG_MODE || defined FAULT_HANDLING_DEBUG_MODE
      printf("martelnk:ring_handler:Last Re-Rx:send_xxxx_token_state> \n\t"
	     "Retransmiting the xxxx Token\n");
#endif
      do{	
	station_possition=(station_possition + 1) % nodes;
	status=read_station(&sta,(station_possition + 1) % nodes);
      }
      while(status==STA_NOT_VALID);

      if(addrcmp(token_packet.token_master_addr,
		 addr_suc_station.ether_addr_octet,
		 ETH_ALEN)){

	/* We check if the highest priority on the token is the       */
	/* token_master.*/

	if(addrcmp(token_packet.addr_prior,
		   addr_suc_station.ether_addr_octet,
		   ETH_ALEN)){
	  
	  addr_suc_station = *(struct ether_addr *)sta.addr_dest;

	  /* In this case we return to the send_read_token state.     */
	  token_master=TRUE;
	  read_token(&token_packet,&Q_tx);
	  status = 
	    write(file_descriptor_error, &token_packet, sizeof(token_packet) );

	  /*status=sendto(socket_srv,&token_packet,sizeof(token_packet),0,NULL,0);*/

	  if(status < 0) 
	    perror("martelnk:send_read_token:failure>sendto_function");
	  
	  last_station_state=send_read_token_state;

	}else{
	  token_packet.packet_id='P';

	  *(struct ether_addr *)token_packet.eth_header.h_dest=
	    *(struct ether_addr *)token_packet.addr_prior;

	  *(struct ether_addr *)token_packet.eth_header.h_source = 
	    current_sta_addr;

	  token_packet.eth_header.h_proto=htons(RT_EP_PROTOCOL);
	  status = 
	    write(file_descriptor_error, &token_packet, sizeof(token_packet) );

 /*status=sendto(socket_srv,&token_packet,sizeof(token_packet),0,NULL,0);*/

 	  if(status < 0) 
	    perror("martelnk:send_permission:failure>sendto_function");  
	  last_station_state=send_permission_state;
	
	}//end if-else(token master_failure))

      }else{
	/* If the failurer succesor isn't the token_master passing    */
	/* the token to the new successor is done.                    */
	
	addr_suc_station = *(struct ether_addr *)sta.addr_dest;

	*(struct ether_addr *)token_packet.eth_header.h_dest = 
	  addr_suc_station;

	status = 
	  write(file_descriptor_error, &token_packet, sizeof(token_packet) );

/*status=sendto(socket_srv,&token_packet,sizeof(token_packet),0,NULL,0);*/
	if(status < 0) perror("martelnk:send_read_token:>sendto_function");
      }// end if -else re-tx.

#if defined MARTELNK_DEBUG_MODE || defined FAULT_HANDLING_DEBUG_MODE
  printf("martelnk:ring_handler>We have redefined the ring.\n");
#endif
      break;

    case send_info_state:

      check_for_valid_station=TRUE;
      token_packet.failure_station=1;
      *(struct ether_addr *) token_packet.failure_station_addr = 
	*(struct ether_addr *)info_packet.eth_header.h_dest;

      /* We inhibit the failure station (which is NOT neccessary our  */
      /* successor)                                                   */
      if(addrcmp(info_packet.eth_header.h_dest,
		 addr_suc_station.ether_addr_octet,
		 ETH_ALEN)){

	nodes=number_nodes();
	inhibit_station(addr_suc_station.ether_addr_octet);
	do{
	  station_possition=(station_possition +1) % nodes;
	  status=read_station(&sta,(station_possition+1) % nodes);
	}
	while(status==STA_NOT_VALID);

	addr_suc_station = *(struct ether_addr *)sta.addr_dest;

      } // end if rx_info_station is successor station
      else {
	inhibit_station(info_packet.eth_header.h_dest);
      }
      token_master=TRUE;

      read_token(&token_packet,&Q_tx);
      status = 
	write(file_descriptor_error, &token_packet, sizeof(token_packet) );
  /* status=sendto(socket_srv,&token_packet,sizeof(token_packet),0,NULL,0);*/
      if(status < 0) perror("martelnk:send_read_token:>sendto_function");
      break;

    case send_permission_state:

      /* We modify the token in order to communicate the others       */
      /* station the failure station.                                 */
 
      check_for_valid_station=TRUE;
      token_packet.failure_station=1;
      *(struct ether_addr *) token_packet.failure_station_addr = 
	*(struct ether_addr *)token_packet.eth_header.h_dest;

      /* We inhibit the failure station (which is NOT neccessary our  */
      /* successor)                                                   */
      if(addrcmp(token_packet.eth_header.h_dest,
		 addr_suc_station.ether_addr_octet,
		 ETH_ALEN)){

	nodes=number_nodes();
	inhibit_station(addr_suc_station.ether_addr_octet);
	status=STA_NOT_VALID;
	
	while(status==STA_NOT_VALID){
	  station_possition=(station_possition +1) % nodes;
	  status=read_station(&sta,(station_possition+1) % nodes);
	}// end while.

	addr_suc_station = *(struct ether_addr *)sta.addr_dest;

      } // end if rx_info_station is successor station
      else {
	inhibit_station(token_packet.eth_header.h_dest);
      }
      token_master=TRUE;

      read_token(&token_packet,&Q_tx);
      status = 
	write(file_descriptor_error, &token_packet, sizeof(token_packet) );
   /*status=sendto(socket_srv,&token_packet,sizeof(token_packet),0,NULL,0);*/
      if(status < 0) perror("martelnk:send_read_token:>sendto_function");
      break;

    default:
      break;
    }//end switch(last_station_state)
    
    retry_counter=0;
  } //end if(retry_counter).

#ifdef FAULT_HANDLING_DEBUG_MODE
  printf("Ha saltado el TIMER: %d veces, y sequence number: %d\n",
	 retry_counter,seq_number); 
#endif
  /*we have to enable again the timer*/
  timerdata.it_interval=timer_disable;
  timerdata.it_value=timeout;
  
  if(timer_settime(timer_id,0,&timerdata,NULL)==-1){
    perror("timer_settimer error\n");
    *ret_val=UNEXPECTED_ERROR;
    pthread_exit(ret_val);
  }

  } // end while(1)

} //end function ring_handler;


#ifdef CPU_OVERHEAD_PROTOCOL_TIMES

#ifndef MAX_CPU_POINTS
#define MAX_CPU_POINTS 50000
#endif
#ifndef CPU_POINTS_DISCARTED
#define CPU_POINTS_DISCARTED 0
#endif
  
int index;
long long tmed_idle=0;
long long tmax_idle=0;
long long tmin_idle=0;
long long tmed_waiting_packet=0;
long long tmax_waiting_packet=0;
long long tmin_waiting_packet=0;
long long tmed_send_read_token=0;
long long tmax_send_read_token=0;
long long tmin_send_read_token=0;
long long tmed_check_token=0;
long long tmax_check_token=0;
long long tmin_check_token=0;
long long tmed_send_permission=0;
long long tmax_send_permission=0;
long long tmin_send_permission=0;
long long tmed_send_token=0;
long long tmax_send_token=0;
long long tmin_send_token=0;
long long tmed_send_info=0;
long long tmax_send_info=0;
long long tmin_send_info=0;
long long tmed_recv_info=0;
long long tmax_recv_info=0;
long long tmin_recv_info=0;

int idle_sta=0;
long long idle_st[MAX_CPU_POINTS];
int waiting_packet=0;
long long waiting_packet_st[MAX_CPU_POINTS];
int send_read_token_sta;
long long send_read_token_st[MAX_CPU_POINTS];
int check_token_sta=0;
long long check_token_st[MAX_CPU_POINTS];
int send_permission_sta=0;
long long send_permission_st[MAX_CPU_POINTS / 2];
int send_token_sta=0;
long long send_token_st[MAX_CPU_POINTS];
int send_info_sta=0;
long long send_info_st[MAX_CPU_POINTS / 2];
int recv_info_sta=0;
long long recv_info_st[MAX_CPU_POINTS / 2];

#endif /* CPU_OVERHEAD_PROTOCOL_TIMES*/


int martelnk(){

  int length_msg, status;
  unsigned char buffer[ETH_FRAME_LEN];
  unsigned short int temp;
  element_t *El;
  struct timespec sleep_time_state;
  int *ret_val;
  pthread_attr_t attr;

/* The file_descriptor */
  int file_descriptor;

#ifdef TOKEN_CHECKING_PRIORITY_TIMES

  struct timespec token_check_prio_first;
  struct timespec token_check_prio_last;

#endif

#ifdef CPU_OVERHEAD_PROTOCOL_TIMES
  
  struct timespec cpu_overhead_first;
  struct timespec cpu_overhead_last;
  long long twait_ns,tidle_temp;

#endif
  
#ifdef FAULT_HANDLING_DEBUG_MODE

#ifndef FAULT_PACKET 
#define FAULT_PACKET 3000
#endif

  struct timespec sleep_time_us = {LISTEN_TIMEOUT_S, LISTEN_TIMEOUT_NS};
  int rx_packet_counter=0;

#endif /*FAULT_HANDLING_DEBUG_MODE*/

/* we use the boolean token_possession to find out which  */
/* station manages the token. */
//this boolean will set the management station.
  boolean token_possession=FALSE; 
   

  enum states station_state=offline_state;

  struct sigevent event;
  pthread_t ring_handler_thread;
  
#ifdef TOKEN_CHECKING_PRIORITY_TIMES

#ifndef MAX_TOKEN_POINTS
#define MAX_TOKEN_POINTS 50000
#endif

  int index1;
  long long token_checking[MAX_TOKEN_POINTS];
  int token_checking_sta=0;
  long long tmax_token_checking=0;
  long long tmin_token_checking=0;
  //  double tmed_token_checking=0.0;
  long long tmed_token_checking=0;
#endif /* TOKEN_CHECKING_PRIORITY_TIMES */



  while(1){
   
   switch(station_state){
     //On Offline state we start the link and configure the stations.
    case offline_state: 
      
#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:offline>Thread on OFFLINE State...\n");
#endif

#ifdef CPU_OVERHEAD_PROTOCOL_TIMES
       clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_first);
#endif


    if( ( file_descriptor = open(DEVICE_NAME, O_RDWR) ) <0){
      perror("Ethernet device open.\n");
      return -1;
    }
    if( ( file_descriptor_error = open(DEVICE_NAME, O_RDWR) ) <0){
      perror("Ethernet device open.\n");
      return -1;
    }
      /*  we attempt to achieve the MAC address of the current station. */

      if( ioctl(file_descriptor, ETH_HWADDR, current_sta_addr.ether_addr_octet)
	  < 0 ){
	perror("ioctl:SIOCGIFHADDR");
	exit(UNEXPECTED_ERROR);
      }
/*    addrcpy(current_sta_addr.sll_addr,ifr.ifr_addr.sa_data,ETH_ALEN); */

#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:martelnk> Current Station: %s \n",
	     ether_ntoa( (struct ether_addr * ) 
			 current_sta_addr.ether_addr_octet ));

#endif

      /* we inicialice the logical ring */
      net_info(&token_possession);


      /*   Now we start the socket */
#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:offline>Creating the communication socket...");
#endif


#ifdef MARTELNK_DEBUG_MODE
      printf("OK!\n");
#endif
      
      /* Promiscuous mode is always set for now.*/
      /* By defaultf the MaRTE OS ethernet driver sets the */

       /*We set up the sleep time of the station to decrease CPU overhead*/
      sleep_time_state.tv_sec=STATE_SLEEP_TIME_S;
      sleep_time_state.tv_nsec=STATE_SLEEP_TIME_NS;

      /* We now prepare the station to receive the timer_timeout signal.*/
       
      event.sigev_notify=SIGEV_SIGNAL;
      event.sigev_signo=SIGRTMAX;

      timeout.tv_sec=LISTEN_TIMEOUT_S;
      timeout.tv_nsec=LISTEN_TIMEOUT_NS;
      timer_disable.tv_sec=0;
      timer_disable.tv_nsec=0;
      if(timer_create(CLOCK_REALTIME, &event, &timer_id)== -1){
	perror("timer creation error\n");
	*ret_val=UNEXPECTED_ERROR;
	pthread_exit(ret_val);
      }

      // Lanzo el thread de ring_handler....
      // Heredar attributos del padre.
      pthread_attr_init(&attr);
      pthread_attr_setinheritsched (&attr, PTHREAD_INHERIT_SCHED);

      if(pthread_create(&ring_handler_thread, &attr, (void *)ring_handler,
			(void *)NULL) != 0){
	printf("martelnk> Error thread creation error.\n");
	exit(UNEXPECTED_ERROR);
      }
      pthread_attr_destroy(&attr);
      /*It is necessary to build a first token. */
      /*the token_possession station will do it.*/
      if(token_possession==TRUE)
	station_state=send_read_token_state;
      else
	station_state=idle_state; 
      
      /*We have to be sure that the error state is set to 0 */
      token_packet.failure_station=0;

#ifdef CPU_OVERHEAD_PROTOCOL_TIMES
      clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_last);   
#endif /* CPU_OVERHEAD_PROTOCOL_TIMES*/

#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:offline>OFFLINE state FINISHED\n");
#endif

      break;

   case idle_state: //On this state the station waits a transmision.
     
#ifdef MARTELNK_DEBUG_MODE
     printf("martelnk:idle>Thread on IDLE State...\n");
#endif

#ifdef CPU_OVERHEAD_PROTOCOL_TIMES
     clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_first);   
#endif

       /*  We inicializate Rx buffer to 0 */
       length_msg=0;
#ifdef MARTELNK_DEBUG_MODE
       printf("martelnk:idle>We begin to listen to any packet...\n");
#endif

       if(check_for_error){
  
#ifdef MARTELNK_DEBUG_MODE
       printf("martelnk:idle>And checking for errors ON\n");
#endif
	 /*we build a non-periodic timer which runs LISTEN_TIMEOUT ns*/ 
	 timerdata.it_interval=timer_disable;
	 timerdata.it_value=timeout;

	 if(timer_settime(timer_id,0,&timerdata,NULL)==-1){
	   perror("timer_settimer error\n");
	   *ret_val=UNEXPECTED_ERROR;
	   pthread_exit(ret_val);
	 }

       }//end if check_for_error.


#ifdef CPU_OVERHEAD_PROTOCOL_TIMES
       clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_last); 
       
       tidle_temp= 
	 ((long long)cpu_overhead_last.tv_sec*1e9 + cpu_overhead_last.tv_nsec) 
	 - ((long long)cpu_overhead_first.tv_sec*1e9 + 
	    cpu_overhead_first.tv_nsec);

       clock_gettime(CLOCK_MONOTONIC,&cpu_overhead_first); 
#endif
       /* We wait (blocking read) to receive a message */
       /* This goto label is necessary in order to recover from duplicated */
       /* packet situation without re-enabling the timeout, thus the timeout */
       /* will be always the same. */
   read_bloquing_section:
       
       length_msg = read(file_descriptor, buffer, ETH_FRAME_LEN );

#ifdef CPU_OVERHEAD_PROTOCOL_TIMES
    
       clock_gettime(CLOCK_MONOTONIC,&cpu_overhead_last); 
       
       twait_ns = 
	 ((long long)cpu_overhead_last.tv_sec*1e9 + cpu_overhead_last.tv_nsec)
	 - ((long long)cpu_overhead_first.tv_sec*1e9 + 
	    cpu_overhead_first.tv_nsec);
       /*We now take the time to meassure the IDLE Stare*/

       clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_first); 
   
#endif

       if(length_msg==-1 && errno==EINTR){
	 station_state=idle_state;

#ifdef MARTELNK_DEBUG_MODE
       printf("martelnk:Idle State:> Error in received packet read "
	      "returned -1\n");
#endif
       //goto read_bloquing_section;
       break; //-> Thus re-enabling the timer.
       } // else (signal ignore?!)

#ifdef MARTELNK_DEBUG_MODE
       printf("martelnk:Idle State:> OK! packet received. %d bytes\n",
	      length_msg);
#endif
   
       /* Since promiscuous mode is set we have to determine */
       /* if the paquet is for us.                           */

       if(memcmp(buffer, current_sta_addr.ether_addr_octet, ETH_ALEN)){
	 /* If it is not for us, we return to idle state. */

#ifdef MARTELNK_DEBUG_MODE
	 printf("martelnk:promiscuous> This packet not for me.It's for: %s\n",
		ether_ntoa((struct ether_addr * )buffer));
#endif
	 station_state=idle_state;
	 
	 if(check_for_error){
	   /* we have to disable the timer */ 
	   timerdata.it_interval=timer_disable;
	   timerdata.it_value=timer_disable;
#ifdef MARTELNK_DEBUG_MODE
       printf("Timer disabled: Received a not for us Packet\n");
#endif
	   if(timer_settime(timer_id,0,&timerdata,NULL)==-1){
	     perror("timer_settimer error\n");
	     *ret_val=UNEXPECTED_ERROR;
	     pthread_exit(ret_val);
	   }
	   /*We also return the retry counter to 0*/ 
	   retry_counter=0;
	   check_for_error=FALSE;

#ifdef MARTELNK_DEBUG_MODE
	   printf("martelnk:idle>And checking for errors OFF\n");
#endif
	 } //end if check_for_error
	 
	 break;
       } // end if the packet is not for me.
       
       temp = (unsigned short int) *(buffer + ETH_HLEN + PACKET_NUMBER_OFFSET);
       
       if(temp == last_packet_number){
	 /* The sequence number is the SAME as the last received */
	 /* packet, so it is discarted.*/
	 station_state=idle_state;
	 
#ifdef FAULT_HANDLING_DEBUG_MODE
	 printf("martelnk:FAULT_HANDLING:> Packet duplicated: %d\n", temp);
#endif
	 // break;
	 /* This goto is needed to return to the read from net function   */
	 /* without re-enabling the timer (in case check_for_error==TRUE) */
	 /* in case duplicated packet is received.                        */
	 goto read_bloquing_section;

       }else{
	 last_packet_number=temp;

	 if(check_for_error){

#ifdef MARTELNK_DEBUG_MODE
       printf("martelnk:Idle State:> Timer disabled: Received a proper "
	      "packet for us.\n");
#endif
	   /*we have to disable the timer*/ 
	   timerdata.it_interval = timer_disable;
	   timerdata.it_value = timer_disable;
	 
	   if(timer_settime(timer_id, 0, &timerdata, NULL)==-1){
	     perror("timer_settimer error\n");
	     *ret_val=UNEXPECTED_ERROR;
	     pthread_exit(ret_val);
	   }
	   /*We also return the retry counter to 0*/ 
	   retry_counter=0;
	   check_for_error=FALSE;
#ifdef MARTELNK_DEBUG_MODE
	   printf("martelnk:Idle State>And checking for errors OFF\n");
#endif
	 } //end if check_for_error
       }// end if-else sequence-number


#ifdef FAULT_HANDLING_DEBUG_MODE

       if(rx_packet_counter == FAULT_PACKET){
	 printf("martelnk:FAULT_HANDLING:>Sleeping %d seg. %d ns\n",
		sleep_time_us.tv_sec,sleep_time_us.tv_nsec);
	 nanosleep(&sleep_time_us,NULL);
	 rx_packet_counter=0;
       }else
	 rx_packet_counter++;
#endif /*FAULT_HANDLING_DEBUG_MODE*/

       /* Now we have received a message, but we don't know whether it is   */
       /* an info packet or a token packet. We can guess it by looking into */
       /* the first byte of the received information, if it is 'T' it's a   */
       /* token packet and if it is 'I' it's an info packet.and if it is 'P'*/
       /* it allows us to tx info. */ 

#ifdef MARTELNK_DEBUG_MODE
	 printf("martelnk:idle>Checking if TOKEN or INFO packet....");
#endif

       switch((char)*(buffer + ETH_HLEN) ){
       case 'T':
	 token_packet=*(token_packet_t *) buffer;
	 station_state=check_token_state;

       
	 /* We have to check if the token has information about a     */
	 /* failure station or if it has done a round informing about */
	 /* a invalid station.                                        */
	 if(check_for_valid_station==TRUE){
	   check_for_valid_station=FALSE;
	   token_packet.failure_station=0;
	 }
       
	 if(token_packet.failure_station!=0){
	   inhibit_station(token_packet.failure_station_addr);
	 }


#ifdef MARTELNK_DEBUG_MODE
	 printf("TOKEN packet ID: %c\n",token_packet.packet_id);
#endif
	 break;
       case 'P':
	 token_packet=*(token_packet_t *) buffer;
	 station_state=send_info_state;
#ifdef MARTELNK_DEBUG_MODE
	 printf("TOKEN packet ID: %c\n",token_packet.packet_id);
#endif
	 break;
       case 'I':
#ifdef MARTELNK_DEBUG_MODE
	 printf("martelnk:idle>Checking if INFO packet....");
#endif
	 info_packet=*(info_packet_t *)buffer;
	 /* If the data sequence number is the one we expect. we continue.*/
	 /* if not we discard that packet.                                */
	   station_state=recv_info_state;
#ifdef MARTELNK_DEBUG_MODE
	 printf("INFO packet\n");
#endif
	 break;
       default:
	   /* if it isn't a token or a info packet we discard it. */	 
	   /* and continue listening                              */
#ifdef MARTELNK_DEBUG_MODE	 
	 printf("martelnk:idle>We have received noise. %c\n",
		token_packet.packet_id);
#endif
	 break;


       } //end switch()


#ifdef CPU_OVERHEAD_PROTOCOL_TIMES
    clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_last); 

    waiting_packet_st[waiting_packet]=twait_ns;
    waiting_packet++;

    idle_st[idle_sta] = 
      ((long long)cpu_overhead_last.tv_sec*1e9 + 
       cpu_overhead_last.tv_nsec)  - 
      ((long long)cpu_overhead_first.tv_sec*1e9 + 
       cpu_overhead_first.tv_nsec) + tidle_temp;

    idle_sta++;

/*     if( !(idle_sta % 50) )  */
/*       printf("idle count %d \n", idle_sta); */

    if(idle_sta==MAX_CPU_POINTS && token_possession){
      

      tmin_idle=idle_st[CPU_POINTS_DISCARTED];
      tmin_send_read_token=send_read_token_st[CPU_POINTS_DISCARTED];
      tmin_check_token=check_token_st[CPU_POINTS_DISCARTED];
      tmin_send_permission=send_permission_st[CPU_POINTS_DISCARTED];
      tmin_send_token=send_token_st[CPU_POINTS_DISCARTED];
      tmin_send_info=send_info_st[CPU_POINTS_DISCARTED];
      tmin_recv_info=recv_info_st[CPU_POINTS_DISCARTED];
      tmin_waiting_packet=waiting_packet_st[CPU_POINTS_DISCARTED];

      for(index=CPU_POINTS_DISCARTED;index<MAX_CPU_POINTS;index++){

	if(idle_st[index]>tmax_idle)
	  tmax_idle=idle_st[index];
	if(idle_st[index]<tmin_idle)
	  tmin_idle=idle_st[index];
	tmed_idle+=idle_st[index]; 

	if(index < waiting_packet){
	  if(waiting_packet_st[index]>tmax_waiting_packet)
	    tmax_waiting_packet=waiting_packet_st[index];

	  if(waiting_packet_st[index]<tmin_waiting_packet)
	    tmin_waiting_packet=waiting_packet_st[index];

	  tmed_waiting_packet += waiting_packet_st[index];
	}

	if(index < send_read_token_sta){
	  if(send_read_token_st[index]>tmax_send_read_token)
	    tmax_send_read_token=send_read_token_st[index];

	  if(send_read_token_st[index]<tmin_send_read_token)
	    tmin_send_read_token=send_read_token_st[index];

	  tmed_send_read_token += send_read_token_st[index];
	}

	if(index < check_token_sta){
	  if(check_token_st[index]>tmax_check_token)
	    tmax_check_token=check_token_st[index];

	  if(check_token_st[index]<tmin_check_token)
	    tmin_check_token=check_token_st[index];

	  tmed_check_token += check_token_st[index];
	}

	if(index < send_permission_sta){
	  if(send_permission_st[index]>tmax_send_permission)
	    tmax_send_permission=send_permission_st[index];

	  if(send_permission_st[index]<tmin_send_permission)
	    tmin_send_permission=send_permission_st[index];

	  tmed_send_permission += send_permission_st[index];
	}


	if(index < send_token_sta){
	  if(send_token_st[index]>tmax_send_token)
	    tmax_send_token=send_token_st[index];

	  if(send_token_st[index]<tmin_send_token)
	    tmin_send_token=send_token_st[index];

	  tmed_send_token += send_token_st[index];
	}

	if(index < send_info_sta){
	  if(send_info_st[index]>tmax_send_info)
	    tmax_send_info=send_info_st[index];

	  if(send_info_st[index]<tmin_send_info)
	    tmin_send_info=send_info_st[index];

	  tmed_send_info+=send_info_st[index];
	}
	
	if(index < recv_info_sta){
	  if(recv_info_st[index]>tmax_recv_info)
	    tmax_recv_info=recv_info_st[index];
	  if(recv_info_st[index]<tmin_recv_info)
	    tmin_recv_info=recv_info_st[index];
	  tmed_recv_info+=recv_info_st[index];
	}

	}// end for.

      /* We calculate the average value outside the for in order to be */
      /* closer to the exact value.                                    */

      tmed_idle /= (idle_sta- CPU_POINTS_DISCARTED);
      tmed_waiting_packet /= (waiting_packet-CPU_POINTS_DISCARTED);
      tmed_send_read_token /= (send_read_token_sta-CPU_POINTS_DISCARTED);
      tmed_check_token /= (check_token_sta - CPU_POINTS_DISCARTED);
      tmed_send_permission /= (send_permission_sta - CPU_POINTS_DISCARTED);
      tmed_send_token /= (send_token_sta-CPU_POINTS_DISCARTED);
      tmed_send_info /= (send_info_sta-CPU_POINTS_DISCARTED); 
      tmed_recv_info /= (recv_info_sta-CPU_POINTS_DISCARTED);

      printf("IDLE State: Tmax: %qd ns, Tmin: %qd ns, Tmed: %qd ns\n",
	     tmax_idle,tmin_idle,tmed_idle);
      printf("SEND READ TOKEN State: Tmax: %qd ns, Tmin: %qd ns, Tmed: %qd ns\n",
	     tmax_send_read_token,tmin_send_read_token,tmed_send_read_token);
      printf("CHECK TOKEN State: Tmax: %qd ns, Tmin: %qd ns, Tmed: %qd ns\n",
	     tmax_check_token,tmin_check_token,tmed_check_token);
      printf("SEND PERMISSION State: Tmax: %qd ns, Tmin: %qd ns, Tmed: %qd ns\n",
	     tmax_send_permission,tmin_send_permission,tmed_send_permission);
      printf("SEND TOKEN State: Tmax: %qd ns, Tmin: %qd ns, Tmed: %qd ns\n",
	     tmax_send_token,tmin_send_token,tmed_send_token);
      printf("SEND INFO State: Tmax: %qd ns, Tmin: %qd ns, Tmed: %qd ns\n",
	     tmax_send_info,tmin_send_info,tmed_send_info);
      printf("RECV INFO State: Tmax: %qd ns, Tmin: %qd ns, Tmed: %qd ns\n",
	     tmax_recv_info,tmin_recv_info,tmed_recv_info);
      printf("WAITING FOR PACKET: Tmax: %qd ns, Tmin: %qd ns, Tmed: %qd ns\n",
	     tmax_waiting_packet,tmin_waiting_packet,tmed_waiting_packet);


      if(token_possession){
	
	printf("Packets sent on IDLE State: %d\n",idle_sta);
	printf("Packets sent on SEND READ TOKEN State: %d\n",
	       send_read_token_sta);
	printf("Packets sent on CHECK TOKEN State: %d\n",check_token_sta);
	printf("Packets sent on SEND_PERMISSION State: %d\n",
	       send_permission_sta);
	printf("Packets sent on SEND_TOKEN State: %d\n",send_token_sta);
	printf("Packets sent on SEND INFO State: %d\n",send_info_sta);
	printf("Packets sent on RECV INFO State: %d\n",recv_info_sta);
	//sleep(10);
	printf("Data taken\n");
      }
    }//end if 
   
    
#endif /* CPU_OVERHEAD_PROTOCOL_TIMES*/

#ifdef MARTELNK_DEBUG_MODE
       printf("martelnk:idle>IDLE state FINISHED\n");
#endif
       break;

    case send_read_token_state:
      /* token_possession = TRUE means that the current station manages    */
      /* the token. So it has to initializate the logical ring by sending  */
      /* a token with information about the highest element in it queue    */
      /* to its sucessor.                                                  */

#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:send_read_token>Thread on SEND_READ_TOKEN State...\n");
#endif

#ifdef CPU_OVERHEAD_PROTOCOL_TIMES
      clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_first); 

#endif

      token_master=TRUE;

#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:send_read_token>I have and rule the TOKEN\n");
      printf("martelnk:send_read_token>Reading the information in order "
	     "to send the token\n");
#endif

      read_token(&token_packet,&Q_tx);

#ifdef TOKEN_CHECKING_PRIORITY_TIMES
      if( clock_gettime(CLOCK_MONOTONIC,&token_check_prio_first) <0)
	perror("send_read_token_state : clock_gettime - "
	       "TOKEN_CHEKING_PRIORITY_TIMES\n");

#endif

      status=write(file_descriptor,&token_packet,sizeof(token_packet));

      if(status < 0){
	perror("martelnk:send_read_token:>sendto_function");
      }

#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:send_read_token>The token has been sent\n");
#endif
      station_state=idle_state;
      last_station_state=send_read_token_state;
      /* And last but not less :) we have to tell the station that has the */
      /* responsability of checking the media for errors                   */
      check_for_error=TRUE;

#ifdef CPU_OVERHEAD_PROTOCOL_TIMES

      clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_last); 


      send_read_token_st[send_read_token_sta] = 
	((long long)cpu_overhead_last.tv_sec*1e9 + cpu_overhead_last.tv_nsec)
	- ((long long)cpu_overhead_first.tv_sec*1e9 + 
	   cpu_overhead_first.tv_nsec);

      send_read_token_sta++;

#endif

#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:send_read_token>SEND_READ_TOKEN state FINISHED\n");
#endif

      break;

    case check_token_state:

#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:check_token>Thread on CHECK_TOKEN State...\n");   
#endif

#ifdef CPU_OVERHEAD_PROTOCOL_TIMES

   clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_first); 
#endif


   /* Comparing the received token information with the information */
   /* in our queue. */

      write_token(&token_packet,&Q_tx);

#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:check_token>We have properly wite_token the token ;-) "
	     "%c\n", token_packet.packet_id);
#endif

#ifdef TOKEN_CHECKING_PRIORITY_TIMES

      if(clock_gettime(CLOCK_MONOTONIC,&token_check_prio_last) <0 )
	perror("check_token_state: clock_gettime - "
	       "TOKEN_CHECKING_PRIORITY_TIMES\n");

      if(token_master==TRUE && token_possession==FALSE){
	token_checking[token_checking_sta] = 
	  ((long long)token_check_prio_last.tv_sec*1e9 + 
	   token_check_prio_last.tv_nsec)  - 
	  ((long long)token_check_prio_first.tv_sec*1e9 + 
	   token_check_prio_first.tv_nsec);

	token_checking_sta++;

	if(token_checking_sta==MAX_TOKEN_POINTS){

	  tmin_token_checking=token_checking[0];

	  for(index1=TOKEN_POINTS_DISCARTED;index1<token_checking_sta;index1++){
	    
	    if(token_checking[index1] > tmax_token_checking)
	      tmax_token_checking = token_checking[index1];

	    if(token_checking[index1] < tmin_token_checking)
	      tmin_token_checking = token_checking[index1];
	     
	    tmed_token_checking += token_checking[index1]; 

	  }// end for.

	  tmed_token_checking = tmed_token_checking / 
	    (token_checking_sta - TOKEN_POINTS_DISCARTED);

	  printf("Token Passing Times: Tmax: %qd ns, Tmin: %qd ns, "
		 "Tmed: %qd ns\n", tmax_token_checking, tmin_token_checking,
		 tmed_token_checking);
	  
	  exit(1);
	} //end if (MAX_TOKEN_POINTS)
	
      }//end if 

#endif


      if(token_master==FALSE || token_packet.priority==0){
	station_state=send_token_state;
      }
      if(token_master==TRUE && token_packet.priority!=0){
	  token_master=FALSE;
	if(addrcmp(token_packet.addr_prior, current_sta_addr.ether_addr_octet,
		   ETH_ALEN)){
	  station_state=send_info_state;
	  /*Since we are transmiting we set token_master to FALSE.*/
	}else
	  station_state=send_permission_state;
      }
#ifdef CPU_OVERHEAD_PROTOCOL_TIMES

      clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_last); 

    check_token_st[check_token_sta] = 
      ((long long)cpu_overhead_last.tv_sec*1e9 + cpu_overhead_last.tv_nsec) - 
      ((long long)cpu_overhead_first.tv_sec*1e9 + cpu_overhead_first.tv_nsec);

    check_token_sta++;
#endif

#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:check_token>CHECK_TOKEN state FINISHED\n");
#endif

      break;

    case recv_info_state:

#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:recv_info>Thread on RECV_INFO State...\n");   
#endif

#ifdef CPU_OVERHEAD_PROTOCOL_TIMES

      clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_first); 
 /*    rdtscll(t0); */
#endif
    /* MODIFICAR!!!!! escribir directamente en element_t*/
/* rx_buffer=calloc(info_packet.length_info,1); */
/* memcpy(rx_buffer,buffer + sizeof(info_packet),info_packet.length_info); */
      write_info(&info_packet, Q_rx, buffer);
      station_state = send_read_token_state;

#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:recv_info>RECV_INFO state FINISHED\n");
#endif
#ifdef CPU_OVERHEAD_PROTOCOL_TIMES

    clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_last); 
/*     rdtscll(t1); */
/*     cycles = (t1-t0-oh); */
/*     us_time=(int)((double)((double)cycles/(double)tps)*1e6); */
/*     printf("RECV_INFO State Overhead: - %5lld - clock ticks - %d - us\n",cycles,us_time); */
    recv_info_st[recv_info_sta] = 
      ((long long)cpu_overhead_last.tv_sec*1e9 + cpu_overhead_last.tv_nsec)
      - ((long long)cpu_overhead_first.tv_sec*1e9 + cpu_overhead_first.tv_nsec);
    recv_info_sta++;
#endif
    break;

   case send_info_state:

#ifdef MARTELNK_DEBUG_MODE
     printf("martelnk:send_info>Thread on SEND_INFO State...\n");   
#endif

#ifdef CPU_OVERHEAD_PROTOCOL_TIMES
     clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_first); 

  /*   rdtscll(t0); */
#endif

    /* We free memory of the last tx info packet.*/
    //    free(tx_buffer);

     read_info(&info_packet,&Q_tx,&El);
   
#ifdef MARTELNK_DEBUG_MODE
     printf("martelnk:send_info>Trying to allocate memory...");
#endif
   
      //tx_buffer=calloc(sizeof(info_packet)+info_packet.length_info,sizeof(char));

#ifdef MARTELNK_DEBUG_MODE
      printf("OK!\n");
      printf("martelnk:send_info>Copying the information to tx...");
#endif

      *(info_packet_t *)tx_buffer=info_packet;
      memcpy(tx_buffer + sizeof(info_packet), El->element_pointer,
	     El->element_size);

#ifdef MARTELNK_DEBUG_MODE
      printf("OK!\n");
      printf("martelnk:send_info>Sending the information...");
#endif

      status = write(file_descriptor, tx_buffer,
		     sizeof(info_packet) + El->element_size );
/*sendto(socket_srv,tx_buffer,sizeof(info_packet)+El.element_size,0,NULL,0); */
      if(status < 0) perror("martelnk:send_info:>sendto_function");
#ifdef MARTELNK_DEBUG_MODE
      printf("OK!\n");
      printf("martelnk:send_info>Unallocating the memory...");
#endif

#ifdef MARTELNK_DEBUG_MODE
      printf("OK!\n");
#endif

      /* we free the memory allocated by the aplication in order to store */
      /* the information. It's no needed anymore. */
      free_element(El); 

      station_state=idle_state;
      last_station_state=send_info_state;
      /* And last but not less :) we have to tell the station that has the */
      /* responsability of checking the media for errors*/
      check_for_error=TRUE;
      
#ifdef CPU_OVERHEAD_PROTOCOL_TIMES

    clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_last); 
    
/*     rdtscll(t1); */
/*     cycles = (t1-t0-oh); */
/*     us_time=(int)((double)((double)cycles/(double)tps)*1e6); */
/*     printf("SEND_INFO State Overhead: - %5lld - clock ticks - %d - us\n",cycles,us_time); */
    send_info_st[send_info_sta] = 
      ((long long)cpu_overhead_last.tv_sec*1e9 + cpu_overhead_last.tv_nsec)
      - ((long long)cpu_overhead_first.tv_sec*1e9 +cpu_overhead_first.tv_nsec);
    send_info_sta++;
#endif

#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:send_info>SEND_INFO state FINISHED\n");
#endif

     break;

    case send_token_state:

      /* If we aren't in the management station and we are not allow to tx */
      /* info we have to give the token to our successor. */

#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:send_token>Thread on SEND_TOKEN State...\n");  
#endif

#ifdef CPU_OVERHEAD_PROTOCOL_TIMES
      clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_first); 
#endif
	
	/* We sleep now in order to decrease CPU overhead */
       if(sleep_time_state.tv_sec !=0 || sleep_time_state.tv_nsec !=0)
	 nanosleep(&sleep_time_state,NULL);
	 
/* status=sendto(socket_srv,&token_packet,sizeof(token_packet),0,NULL,0); */
      status=write(file_descriptor,&token_packet,sizeof(token_packet));
      if(status < 0) perror("martelnk:send_token:>sendto_function");

      station_state=idle_state;
      last_station_state=send_token_state;
      /* And last but not less :) we have to tell the station that has the */
      /* responsability of checking the media for errors*/
      check_for_error=TRUE;
            

#ifdef CPU_OVERHEAD_PROTOCOL_TIMES

      clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_last);

      send_token_st[send_token_sta]=
	((long long)cpu_overhead_last.tv_sec*1e9 + cpu_overhead_last.tv_nsec)
	- ((long long)cpu_overhead_first.tv_sec*1e9 + 
	   cpu_overhead_first.tv_nsec);
      send_token_sta++;
#endif
#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:send_token>SEND_TOKEN state FINISHED\n");
#endif

      break;

    case send_permission_state:

#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:send_permission>Thread on SEND_PERMISSION State...\n"); 
#endif

#ifdef CPU_OVERHEAD_PROTOCOL_TIMES
   clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_first); 

#endif

      token_packet.packet_id='P';

      *(struct ether_addr *)token_packet.eth_header.h_dest =
	*(struct ether_addr *)token_packet.addr_prior;
      *(struct ether_addr *)token_packet.eth_header.h_source=current_sta_addr;

      token_packet.eth_header.h_proto=htons(RT_EP_PROTOCOL);

      status=write(file_descriptor,&token_packet,sizeof(token_packet));

      if(status < 0) perror("martelnk:send_read_token:>sendto_function");
      station_state=idle_state;
      last_station_state=send_permission_state;
      /* And last but not less :) we have to tell the station that has the */
      /* responsability of checking the media for errors*/
      check_for_error=TRUE;

#ifdef MARTELNK_DEBUG_MODE
      printf("martelnk:send_permission>SEND_PERMISSION state FINISHED\n");
#endif


#ifdef CPU_OVERHEAD_PROTOCOL_TIMES

    clock_gettime(CLOCK_THREAD_CPUTIME_ID,&cpu_overhead_last); 

    send_permission_st[send_permission_sta] =
      ((long long)cpu_overhead_last.tv_sec*1e9 + cpu_overhead_last.tv_nsec)
      - ((long long)cpu_overhead_first.tv_sec*1e9 + cpu_overhead_first.tv_nsec);

/*     printf("martelnk:send_permission>SEND_PERMISSION state FINISHED: %qd\n", */
/* 	   send_permission_st[send_permission_sta]); */
    send_permission_sta++;
#endif

      break;

    }//end switch.

  }//end while.

  return 0;
}

