/*----------------------------------------------------------------------------
 *-------------------------         R T - E P         ------------------------
 *----------------------------------------------------------------------------
 *                                                             {MARTE_VERSION}
 *
 *                            'm a r t e l n k . h'
 *
 *                                     C
 *
 *                                               
 * File 'martelnk.h'                                                  By Chema.
 *                                                          Jose Maria Martinez
 *                                                         <martinjm@unican.es>
 *
 * Type & main thread declarations.
 *
 *
 *---------------------------------------------------------------------------*/
 /*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */


#ifndef FILE_MARTELNK_H
#define FILE_MARTELNK_H
#include "prot_types.h"
#include "prio_priority.h"
#include <net/if_ether.h>

/**********************************************************************/
/*  History list ******************************************************/
/**********************************************************************/
/*                                                                    */
/*                                                                    */
/*  - Version 0.5                                                     */
/*    MaRTE OS Implementation + several code optimization             */
/*  - Version 0.3a                                                    */
/*    Added error handling.                                           */
/*  - Version 0.2a                                                    */
/*    Promiscuous mode in order to being able to detect               */
/*    errors by means of hearing the channel                          */
/*  - Version 0.1a                                                    */
/*    Added PACKET_SOCK funcionality                                  */
/*  - Version 0.0                                                     */
/*    Protocol implementation with UDP packets                        */
/**********************************************************************/

#define VERSION "0.5"

/* BUFFER stores the Ethernet frame length minus preamble and FCS     */
/*  which makes a total of 1514 bytes.                                */
/* #define BUFFER ETH_FRAME_LEN */
/* MAX_INFO_LENGTH stores the maximun length of the higher layers     */
/*  information packets                                               */
/*#define MAX_INFO_LENGTH ETH_FRAME_LEN - sizeof(info_packet_t)*/
/* #define PROTOCOL ETH_P_ALL */
/*#define PROTOCOL 0x1000*/


/**********************************************************************/
/* Types definition. **************************************************/ 
/**********************************************************************/

/* info_packet_t is the header of the information packet.             */

typedef struct {
  struct ethhdr eth_header;
  unsigned char packet_id;
  priority_t priority;
  unsigned short int packet_number;
  positive_t channel_id;
  positive_t length_info;
} info_packet_t;

/* token_packet_t is the token we will use in the token passing       */
/* transmision. Priority 0 is reserved to maintenance purposes and    */
/* cannot be used.                                                    */

typedef struct {
  struct ethhdr eth_header;
  unsigned char packet_id;
  priority_t priority;
  unsigned short int packet_number;
  unsigned char token_master_addr[ETH_ALEN];
  positive_t failure_station;
  unsigned char failure_station_addr[ETH_ALEN];
  unsigned char addr_prior[ETH_ALEN];
} token_packet_t;

/* We define the different offsets from the RT-EP token / info packet*/

#define PACKET_NUMBER_OFFSET   2 


/* martelnk is the thread function in charge of implementing RT-EP    */
/* protocol. It implement the bus access arbitration                  */
extern int martelnk();

#endif /*FILE_MARTELNK_H*/
