/*----------------------------------------------------------------------------
 *-------------------------         R T - E P         ------------------------
 *----------------------------------------------------------------------------
 *                                                             {MARTE_VERSION}
 *
 *                        'p r i o _ e l e m e n t . c'
 *
 *                                     C
 *
 *                                               
 * File 'prio_element.c'                                              By Chema.
 *                                                          Jose Maria Martinez
 *                                                         <martinjm@unican.es>
 *
 * Body function from prio_element.h.
 *
 *
 *---------------------------------------------------------------------------*/
 /*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sched.h>
#include <pthread.h>
#include "prot_types.h"
#include "prio_element.h"


static boolean inicializated=FALSE;

static element_t *free_memory_elements[MAX_MEMORY_ELEMENT_TYPE];

static unsigned int current_free_memory_element;

/* We need a mutex to assign the free memory area. */
static pthread_mutex_t the_mutex;

boolean equal_element(element_t *str1,element_t *str2){

  if (memcmp((const void *)str1->element_pointer,
	     (const void *)str2->element_pointer, str1->element_size) == 0)
    return TRUE;
  else
    return FALSE;

} //end function equal_element

int element_monitor_init(int prio_ceiling){

  pthread_mutexattr_t attr;

  if(inicializated)
    return 0;
  if(pthread_mutexattr_init(&attr)!=0)
    return CREATION_ERROR;
  if(pthread_mutexattr_setprotocol(&attr,PTHREAD_PRIO_PROTECT)!=0)
    return CREATION_ERROR;
  if(pthread_mutexattr_setprioceiling(&attr,prio_ceiling)!=0)
    return CREATION_ERROR;
  if(pthread_mutex_init(&the_mutex,&attr)!=0)
    return CREATION_ERROR;
  if(pthread_mutexattr_destroy(&attr)!=0)
    return CREATION_ERROR;

  current_free_memory_element=0;
  inicializated=TRUE;
  return 0;
}


element_t *alloc_element(){

  element_t *mem_alloc;

  if(pthread_mutex_lock(&the_mutex)!=0)
    return NULL;
  
  if(current_free_memory_element==0){
    mem_alloc = (element_t *)malloc(sizeof(element_t));    
  }else{
    current_free_memory_element--;
    mem_alloc = free_memory_elements[current_free_memory_element];
  }
  if(pthread_mutex_unlock(&the_mutex)!=0)
    return NULL;
  return mem_alloc;

}//end function alloc_element

int free_element(element_t *E){

  if(pthread_mutex_lock(&the_mutex)!=0)
    return SYNCH_ERROR;
  if(current_free_memory_element==MAX_MEMORY_ELEMENT_TYPE){
    printf("prio_element:element_free:> Too much memory to remmember:"
	   " Wasting memory\n");
  if(pthread_mutex_unlock(&the_mutex)!=0)
      return SYNCH_ERROR;
  return MEM_LOST;
  }
  free_memory_elements[current_free_memory_element] = E;
  current_free_memory_element++;
  if(pthread_mutex_unlock(&the_mutex)!=0)
    return SYNCH_ERROR;
  return 0;
}//end function free_element

