/*----------------------------------------------------------------------------
 *-------------------------         R T - E P         ------------------------
 *----------------------------------------------------------------------------
 *                                                             {MARTE_VERSION}
 *
 *                        'p r i o _ e l e m e n t . h'
 *
 *                                     C
 *
 *                                               
 * File 'prio_element.h'                                              By Chema.
 *                                                          Jose Maria Martinez
 *                                                         <martinjm@unican.es>
 *
 * Type element and functions.
 *
 *
 *---------------------------------------------------------------------------*/
 /*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */

#ifndef FILE_PRIO_ELEMENT_H
#define FILE_PRIO_ELEMENT_H
#include <stdio.h>
#include <string.h>
#include <net/if_ether.h>
#include "prot_types.h"


#define MEM_LOST -40
/************************************************************/
/* element_t is the element of info. The application MUST   */
/* fullfill this stucture in order to Tx information:       */ 
/* - element_pointer is a pointer to the information to     */
/*   transmit                                               */
/* - element_size is the size of the element to tx.         */
/* - addr_dest is an array which has to store the receiber  */
/*   station                                                */
/* - channel_id is the identifier of the communication      */
/*   endpoint                                               */
/************************************************************/
/* element_t : if the type that all the protocol will use to */
/*             carry the message information. This type MUST */
/*             be used with the supplied functions and always*/
/*             by reference that is you always MUST declare  */
/*             as a POINTER. And you must alloc memory with  */
/*             element_alloc()                               */

typedef struct {
  char element_pointer[MAX_INFO_LENGTH];
  positive_t element_size;
  unsigned char addr_dest[ETH_ALEN];
  positive_t channel_id;
} element_t;


/* element_monitor_init : Will initializate the mutex and    */
/*                        some globals                       */
/*                        Returns 0 if OK.                   */
/*                        CREATION_ERROR on error.           */
extern int element_monitor_init(int prio_ceiling);

/* alloc_element : Will return a element_t reserved area     */
/*                 pointer                                   */
/*                 Returns NULL if error                     */
extern element_t *alloc_element();

/* free_element : Will "liberate" a reserved element_t, this */
/*                liberated pointer will be ready to be used */
/*                by another thread calling element_alloc    */
/*                Returns 0 if OK.                           */
/*                MEM_LOST if the buffer of avaiable free    */
/*                memory positions get exhausted.            */
extern int free_element(element_t *E);

extern boolean equal_element(element_t *str1, element_t *str2);
/* returns TRUE if equal.                                   */

#endif // FILE_PRIO_ELEMENT_H

