/*----------------------------------------------------------------------------
 *-------------------------         R T - E P         ------------------------
 *----------------------------------------------------------------------------
 *                                                             {LINUX_VERSION}
 *
 *                           'p r i o _ q u e u e . c'
 *
 *                                     C
 *
 *                                               
 * File 'prio_queue.c'                                                By Chema.
 *                                                          Jose Maria Martinez
 *                                                         <martinjm@unican.es>
 *
 * Body function from prio_queue.h.
 *
 *
 *---------------------------------------------------------------------------*/
 /*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */
#include <stdio.h>
#include <stdlib.h>
#include "prio_queue.h"



/* fuction dynamic_array will help us to access a dynamic "array" */
/* The use is similar to an array, *p is a pointer to the data structure */
/* (the name of the array) and the integer position is the element of */
/* the array, i.e. pos=0 means we are accesing to the first element  */
/* of the "array" which is the same element that points *p. */
cell_t *dynamic_array(cell_t *p, positive_t pos ){

  return p+(pos*sizeof(cell_t));

}//end function dynamic_array.


void upheap(positive_t N, queue_t *Q){

  positive_t p;
  positive_t k=N;
  cell_t temp, *cell_k, *cell_p;

  while(k >= 2){
    p=k/2;
    cell_k=dynamic_array(Q->Q,k);
    cell_p=dynamic_array(Q->Q,p);
    if(cell_k->Pri > cell_p->Pri){

      temp=*cell_k;
      *cell_k=*cell_p;
      *cell_p=temp;
      k=p;
    } //end if
      else
	break;
    //end if- else.
  } //end while.
} // end function unheap.

void init_queue(queue_t *Q){
  Q->Length=0;
  Q->Q=NULL;

}//end function init

boolean empty(queue_t *Q){

  return Q->Length==0;

}//end fuction empty.

positive_t length_queue(queue_t *Q){

  return Q->Length;

}// en function length_queue.

int enqueue(element_t *E, priority_t P, queue_t *Q){

  if (Q->Q==NULL){
    
    Q->Q=(cell_t *)calloc(PRIO_INITIAL_SIZE,sizeof(cell_t));
    Q->queue_size=PRIO_INITIAL_SIZE;

  } // end if.
  else{
    if ((Q->Length +1) >= Q->queue_size){
     printf("prio_queue : enqueue:> Priority queue FULL \n ");
     return(-1);
      // Q->queue_size *=2;
     // Q->Q=(cell_t *)realloc((void *)Q->Q,Q->queue_size*sizeof(cell_t));
      /* since realloc doesn't initialize the new memory area, we use */
      /* memset to "clean" the new area. */
     // memset(dynamic_array(Q->Q,Q->Length + 1),0,(Q->queue_size - Q->Length)*sizeof(cell_t));
    } //end if within else.
  } //end else.
  Q->Length += 1;
  dynamic_array(Q->Q,Q->Length)->E=E;
  dynamic_array(Q->Q,Q->Length)->Pri=P;
 
  upheap(Q->Length,Q);
  return 0;
}//end function enqueue.

void downheap(positive_t K, positive_t N, queue_t *Q){
  
  positive_t j=2*K;
  positive_t half=K;
  cell_t temp=*dynamic_array(Q->Q,K);

  while(j<=N){
    if ( j < N )
      if ( (dynamic_array(Q->Q,j+1)->Pri) > (dynamic_array(Q->Q,j)->Pri) ) 
	j++;
    if ( !( (dynamic_array(Q->Q,j)->Pri) > (temp.Pri) ) ){

      break;
    }// end if
    *dynamic_array(Q->Q,half)=*dynamic_array(Q->Q,j);

    half=j;
    j*=2;
  }// end while.
  *dynamic_array(Q->Q,half)=temp;

}//end function downheap.

int dequeue(element_t **E, priority_t *P, queue_t *Q){

  cell_t *temp_1=dynamic_array(Q->Q,1);
  cell_t *temp_Length=dynamic_array(Q->Q,Q->Length);
    
  if(Q->Length==0)
    return QUEUE_EMPTY; 

  *E=temp_1->E;
  *P=temp_1->Pri;
  
  *temp_1=*temp_Length; 

  Q->Length=Q->Length - 1;

  downheap(1,Q->Length,Q);
  return 0;
}//end function dequeue.




int read_first(element_t **E, priority_t *P, queue_t *Q){
  
  cell_t *temp_1=dynamic_array(Q->Q,1);
  if( Q->Length==0)
    return QUEUE_EMPTY;
  else{
    *E=temp_1->E;
    *P=temp_1->Pri;

  }//end if-else clause.
  return 0;
}//end function read_first.



