/*----------------------------------------------------------------------------
 *-------------------------         R T - E P         ------------------------
 *----------------------------------------------------------------------------
 *                                                             {MARTE_VERSION}
 *
 *                  'p r i o _ q u e u e _ m o n i t o r . h'
 *
 *                                     C
 *
 *                                               
 * File 'prio_queue_monitor.h'                                        By Chema.
 *                                                          Jose Maria Martinez
 *                                                         <martinjm@unican.es>
 *
 * Queue access monitor implementation.
 *
 *
 *---------------------------------------------------------------------------*/
 /*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */


/* Header of the monitor which allow us to access */
/* in a syncronized way to the queue */
#ifndef FILE_PRIO_QUEUE_MONITOR_H
#define FILE_PRIO_QUEUE_MONITOR_H

#include "prot_types.h"
#include "prio_queue.h"

#define TX_QUEUE 0

/* We define the transmision queue and the MAX_CHANNELS     */
/* reception queue as global variables in order to be       */
/* seen by our transmision thread.                          */
queue_t Q_tx,Q_rx[MAX_CHANNELS+1];


/* You can monitor two queues at the same time by specifing the queue number */
/* (queue_num) which might be 0 for the Tx queue and any number for the Rx queue */

extern int create_queue(int prio_ceiling, queue_t *Q, const int queue_num);
/* create_queue must finish before the object is used */
/* create_queue returns 0 if OK, or CREATION_ERROR if */
/* an error occurs */


/* All these functions writes/handles with element_t pointers, the argument MUST*/
/* be a element_t pointer allocated with alloc_element() (queue_insert) except  */
/* the returning-value functions.(queue_extract & read_first_in_queue)          */
/* When the element is no needed anymore you have to use the free_element       */
/* function                                                                     */

extern int queue_insert(element_t *E, priority_t P,queue_t *Q, const int queue_num);
/* queue_insert returns 0 if OK or SYNCH_ERROR */

extern int queue_extract(element_t **E,priority_t *P,queue_t *Q, const int queue_num);
/* queue_extract suspends the calling task until data is avaible */
/* returns 0 if OK or SYNCH_ERROR */

extern int read_first_in_queue(element_t **E, priority_t *P, queue_t *Q,const int queue_num);
/* read_first_in_queue reads, and only reads (doesn't extract)*/
/* the pointer to first element of the queue. Returns 0 if OK.*/
/* On error it returns QUEUE_EMPTY and SYNCH_ERROR */  


extern int is_queue_empty(element_t *E,queue_t *Q,const int queue_num);
/* is_queue_empty checks the queue if there is any element */
/* Returns: 0 if there is any and 1 if there isn't On error it returns */
/* QUEUE_EMPTY and SYNCH_ERROR */  

#endif /*FILE_PRIO_QUEUE_MONITOR_H*/
