/*----------------------------------------------------------------------------
 *-------------------------         R T - E P         ------------------------
 *----------------------------------------------------------------------------
 *                                                             {MARTE_VERSION}
 *
 *                          'p r o t _ t y p e s . h'
 *
 *                                     C
 *
 *                                               
 * File 'prot_types.h'                                               By Chema.
 *                                                          Jose Maria Martinez
 *                                                         <martinjm@unican.es>
 *
 * RT-EP Configuration package. 
 *
 *
 *---------------------------------------------------------------------------*/
 /*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */

/***************************************************************/
/********************** PROT_TYPES.H ***************************/
/***************************************************************/

/* This file is used to define some useful constant,functions  */
/* and to properly configure the protocol and the logical ring.*/


/***************************************************************/
/* We may define here some useful types.                       */
#ifndef FILE_PROT_TYPES_H
#define FILE_PROT_TYPES_H

#include "rt_ep_errorn.h"

/***************************************************************/
// we are atempting the protocol in Linux
//#define LINUX_STATE
// If we are in Marte uncomment the following and comment the previous
#define MARTE_STATE

/*******************************************************************/
/* PRIO_INITIAL_SIZE : Is the size in cells of the reception queue */
/*                     So it is the number of messages that will   */
/*                     store each reception queue and Tx queue.    */
/*                     Since realloc isn't defined in MaRTE, a     */
/*                     queue mustn't have more than                */
/*                     PRIO_INITIAL_SIZE elements.                 */

#define PRIO_INITIAL_SIZE 256

/*************************************************************/
/*  MAX_CHANNELS is the ceil  of the number of               */
/*  Tasks that can be executed in the real-time system       */
/*  you have to adjust it to the proper number               */
/*  The Tasks have to be numered in '1..MAX_CHANNELS' style. */
#define MAX_CHANNELS 5 


/****************************************************************/
/*  MAX_MEMORY_ELEMENT_TYPE is the ceil of element_t types that */
/*  can be managed by the protocol. This is needed in order to  */
/*  allocate memory for the incomming/outgoing messages.        */
/*  It is scalled by one more channel to give some freedom to   */
/*  the protocol in using the type element_t                    */

#define MAX_MEMORY_ELEMENT_TYPE ((MAX_CHANNELS+1)*PRIO_INITIAL_SIZE)


/*******************************************************************/
/* FILE_NAME_CONFIG  Will set the file that the protocol will read */
/*                   to get the ring configuration.                */
/*   File structure:                                               */
/*                                                                 */
/*   mac_address1                                                  */
/*   mac_address2                                                  */
/*  i.e...                                                         */
/*   00:4F:4C:02:F6:68                                             */
/*   00:4F:49:04:7D:8D                                             */
/*******************************************************************/
//#define FILE_NAME_CONFIG "./lnk.conf"

/*******************************************************************/
/* MAX_STATIONS will set the upper limit of the munber of stations */
/* which are going to get into the logical ring                    */
#define MAX_STATIONS 100



/********************************************************************/
/*  Here we define the name of the device which is going to be used */ 
#define DEVICE_NAME "/dev/eth0"


/*************************************************************/
/* MAX_INFO_LENGTH  ******************************************/
/* MAX_INFO_LENGTH stores the maximun length of the higher   */
/*                 layers information packets                */
/*************************************************************/
#define MAX_INFO_LENGTH 1492 

/*************************************************************/
/* RT_EP_PROTOCOL ********************************************/
/* Will indicate the value associated to this protocol.      */

#define RT_EP_PROTOCOL 0x1000


/**************************************************************/
/* STATE_SLEEP_TIME *******************************************/
/* This will define the time the station will sleep in order  */
/* to reduce the CPU Overhead                                 */
/* The units of STATE_SLEEP_TIME_NS are in nanoseconds.!! and */
/*    has to be < 1e9 !!!                                     */
/* The units of STATE_SLEEP_TIME_S are in seconds.            */

#define STATE_SLEEP_TIME_S 0
#define STATE_SLEEP_TIME_NS 1000

/**************************************************************/
/* FAULT HANDLING. ********************************************/
/**************************************************************/
/* LISTEN_TIMEOUT defines the time a station is listening the */
/* media to check if token or data loss.                      */
/* The units of LISTEN_TIMEOUT_NS are in nanoseconds.!! and   */
/*    has to be < 1e9 !!!                                     */
/* The units of LISTEN_TIMEOUT_S are in seconds.              */
/* MAX_RETRIES defines the number of reties the protocol will */
/* attempt before discarting the possible failure station.    */

//#define LISTEN_TIMEOUT_S 4
//#define LISTEN_TIMEOUT_NS 90000000
//#define MAX_RETRIES 20
#define LISTEN_TIMEOUT_S 30
#define LISTEN_TIMEOUT_NS 90000000
#define MAX_RETRIES 20

/**************************************************************/
/* PROTOCOL DEBUG MODE ****************************************/
/**************************************************************/
/* With these defines uncommented you should enter to a       */
/* verbose mode of the protocol modules.                      */
/* FAULT_HANDLING_DEBUG_MODE -> The receiver station will     */
/*                              sleep LISTEN_TIMEOUT seconds  */
/*                              when received FAULT_PACKET    */
/*                              frames are processed.*/

//#define MARTELNK_DEBUG_MODE 
//#define FAULT_HANDLING_DEBUG_MODE
//#define FAULT_PACKET 3000 
//#define MONITOR_DEBUG_MODE 
//#define LNK_CONFIG_DEBUG_MODE
//#define RT_COMM_DEBUG_MODE
/*************************************************************/
/* With these defines uncommented you should enter onto a CPU time*/
/* measure mode. You sould uncomment one define each time */
/* and obviously you will have to comment the debug messages.*/
/* - MAX_POINTS will set the number of frames do you want to be*/
/* taken in the stadistics minimun 50.*/


//#define CPU_OVERHEAD_PROTOCOL_TIMES
//#define MAX_CPU_POINTS 500000
//#define CPU_POINTS_DISCARTED 0


/*************************************************************/
/* With these defines uncommented you should enter onto a TOKEN time*/
/* measure mode. You sould uncomment one define each time */
/* and obviously you will have to comment the debug messages.*/

//#define TOKEN_CHECKING_PRIORITY_TIMES
//#define MAX_TOKEN_POINTS 2000
//#define TOKEN_POINTS_DISCARTED 0

/************************************************************/
/* We define here some "constant utilities" */
#define TRUE 1
#define FALSE 0
 // #define UNEXPECTED_ERROR -55


/*************************************************************/
/* Here we set some useful types.*/
typedef char boolean; // boolean type

typedef unsigned short int positive_t; //tipo positivo 2 bytes

/* we will use addr_cmp function to compare addresses it return TRUE if equal*/
/* and FALSE if not.. it's much faster than memcmp*/

extern int addrcmp(unsigned char *addr1, unsigned char *addr2,const int len);

/* This function copies the address from one array to other. It's much faster *//* than memcpy */

extern int addrcpy(unsigned char *addr1, unsigned char *addr2, const int len);

#endif /*FILE_PROT_TYPES_H*/ 

