/*----------------------------------------------------------------------------
 *-------------------------         R T - E P         ------------------------
 *----------------------------------------------------------------------------
 *                                                             {MARTE_VERSION}
 *
 *                             'r t _ c o m m . c'
 *
 *                                     C
 *
 *                                               
 * File 'rt_comm.c'                                                   By Chema.
 *                                                          Jose Maria Martinez
 *                                                         <martinjm@unican.es>
 *
 * Body functions from rt_comm.h.
 *
 *
 *---------------------------------------------------------------------------*/
 /*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */


#include <pthread.h>
#include "rt_comm.h"
#include "prot_types.h"
#include "martelnk.h"
#include "lnk_config.h"
#include "prio_element.h"
#include "prio_queue_monitor.h"

#define TX_QUEUE 0

boolean comm_init=TRUE; 


/* We use this global variable in order to be sure that the */
/* link is only inicializated once.                         */

int send_info(char *addr_dest, unsigned short int channel_id, 
	      char *info, unsigned short int len, unsigned char priority){

  int status;
  element_t *E;
#ifdef RT_COMM_DEBUG_MODE
  printf("rt_comm:send_info> Starting.....\n ");

#endif
  /* Cheching if a valid station.                           */
  if( (status=check_station(addr_dest)) < 0){
     return status;
  }
  /* Checking if message too long                           */
  if( len > MAX_INFO_LENGTH )
    return INFO_LENGTH_OVERFLOW;

  /* Checking if invalid priority, priority 0 is used only  */
  /* for protocol internal use.                             */

  if(priority==0)
    return INVALID_PRIORITY;

  /* Checking if invalid channel                            */
  if(channel_id == 0 || channel_id > MAX_CHANNELS  )
    return INVALID_CHANNEL;

  E=alloc_element();
  memcpy(E->element_pointer, info, len);
  memcpy(E->addr_dest, addr_dest, ETH_ALEN);
  E->element_size=len;
  E->channel_id=channel_id;

#ifdef RT_COMM_DEBUG_MODE
  printf("rt_comm:send_info:> Sending....\n");
#endif
/*   printf("rt_comm:send_info:> Mandado: Funcion: %d, Info: %d\n",  */
/* 	 *(unsigned char *)E->element_pointer,  */
/* 	 *(int *)(E->element_pointer+sizeof(unsigned char))); */
#ifdef RT_COMM_DEBUG_MODE
      printf("rt_comm:send_info> Dest_Addr: %X:%X:%X:%X:%X:%X ; Channel: %d\n "
	     "rt_comm:send_info> Info: %s ; Size: %d \n"
	     "rt_comm:send_info> Priority: %d \n",addr_dest[0],
	     addr_dest[1],addr_dest[2],addr_dest[3],addr_dest[4],addr_dest[5],
	     channel_id,E->element_pointer,E->element_size, priority);
#endif
 
  switch(queue_insert(E,priority,&Q_tx,TX_QUEUE)){
  case SYNCH_ERROR:
    printf("rt_comm:send_info> Synch error in monitor\n");
    exit(SYNCH_ERROR);
    break;
  case 0:
    return 0;
    // All is OK.
    break;
  default:
    printf("rt_comm:send_info> Unexpected error.\n");
    exit(UNEXPECTED_ERROR);
    break;
  }//end switch

}//end function send_info.


int recv_info(char *addr_sourc, unsigned short int channel_id, 
		     char *recv_info, unsigned char *priority){

  element_t *E;
  unsigned int len;

#ifdef RT_COMM_DEBUG_MODE
  printf("rt_comm:recv_info> Starting.....\n ");
#endif


  /* Checking if invalid channel                            */

  if(channel_id == 0 || channel_id > MAX_CHANNELS  )
    return INVALID_CHANNEL;

#ifdef RT_COMM_DEBUG_MODE
  printf("rt_comm:recv_info> Waiting for a packet.\n ");
#endif    
  switch(queue_extract(&E, priority, &Q_rx[channel_id], channel_id)){
  case SYNCH_ERROR:
    printf("rt_comm:recv_info> Synch error in monitor\n");
    exit(SYNCH_ERROR);
    break;
  case 0:
    break;
  default:
    printf("rt_comm:recv_info> Unexpected error.\n");
    exit(UNEXPECTED_ERROR);
    break;
  }//end switch

  memcpy(recv_info, E->element_pointer, E->element_size);
  memcpy(addr_sourc, E->addr_dest, ETH_ALEN);
  len = E->element_size;
  free_element(E);
  //jlxx  printf("rt_comm:recv_info:> Receiving....\n");
/*   printf("rt_comm:recv_info:> Recibido: Funcion: %d, Info: %d\n", *(unsigned char *)E->element_pointer, *(int *)(E->element_pointer+sizeof(unsigned char))); */
#ifdef RT_COMM_DEBUG_MODE
      printf("rt_comm:recv_info> Sour_Addr: %X:%X:%X:%X:%X:%X ; Channel: %d\n "
	     "rt_comm:recv_info> Info: %s ; Size: %d \n"
	     "rt_comm:recv_info> Priority: %d \n",addr_sourc[0],
	     addr_sourc[1],addr_sourc[2],addr_sourc[3],addr_sourc[4],
	     addr_sourc[5],channel_id, E->element_pointer, E->element_size,
	     *priority);
#endif

  return len;
} //end function recv_info


int any_info(unsigned int channel_id ){

  element_t *E;
 /* Checking if invalid channel                            */
  if(channel_id == 0 || channel_id > MAX_CHANNELS  )
    return INVALID_CHANNEL;

  switch(is_queue_empty(E,&Q_rx[channel_id],channel_id)){
  case SYNCH_ERROR:
    printf("rt_comm:recv_info> Synch error in monitor\n");
    exit(SYNCH_ERROR);
    break;
  case 0:
    return 0;
    break;
  case 1:
    return NO_ELEMENTS;
    break;
  default:
    printf("rt_comm:recv_info> Unexpected error.\n");
    exit(UNEXPECTED_ERROR);
    break;
  }//end switch


}//end function any_info



int init_comm(int prio_ceiling){

  pthread_t martelnk_thread;
  positive_t i;
  int status;
  pthread_attr_t attr;
  struct sched_param param;

  if(comm_init){
#ifdef RT_COMM_DEBUG_MODE
    printf("rt_comm:init_comm> Trying to init. the queues.\n");
    
    printf("rt_comm:init_comm> Tx queue ... ");
#endif
    if((status=create_queue(prio_ceiling,&Q_tx,TX_QUEUE)) <0 )
    	return status; 
    	
#ifdef RT_COMM_DEBUG_MODE
    printf("OK!\n");
#endif
    
    for(i=1;i<=MAX_CHANNELS;i++){
#ifdef RT_COMM_DEBUG_MODE
      printf("rt_comm:init_comm> Rx queue number: %d ... ",i);
#endif
      if((status=create_queue(prio_ceiling,&Q_rx[i],i))<0)
      	return status;
#ifdef RT_COMM_DEBUG_MODE
    printf("OK!\n");
#endif
    }//end for.

   
#ifdef RT_COMM_DEBUG_MODE
      printf("rt_comm:init_comm> Element monitor init...");
#endif
      if((status=element_monitor_init(prio_ceiling)) < 0)
	return status;
#ifdef RT_COMM_DEBUG_MODE
    printf("OK!\n");
#endif
    pthread_attr_init(&attr);
    /* We set the prio ceiling of the Main Communication Thread. */
    param.sched_priority=prio_ceiling;

    pthread_attr_setschedparam(&attr,&param);

    if(pthread_create(&martelnk_thread,&attr,(void *)martelnk,
		      (void *)NULL) != 0){
      printf("rt_comm.c:init_comm> Main thread creation error.\n");
      exit(UNEXPECTED_ERROR);
    }
    pthread_attr_destroy(&attr);
    comm_init=FALSE; 
  }
  return 0;	
}// end function init_comm.


