/*----------------------------------------------------------------------------
 *-------------------------         R T - E P         ------------------------
 *----------------------------------------------------------------------------
 *                                                             {MARTE_VERSION}
 *
 *                             'r t _ c o m m . h'
 *
 *                                     C
 *
 * 
 * File 'rt_comm.h'                                                By Chema.
 *                                                        Jose Maria Martinez
 *                                                       <martinjm@unican.es>
 *
 * Application interface.
 *
 *
 *---------------------------------------------------------------------------*/
 /*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */


#ifndef FILE_RT_COMM_H
#define FILE_RT_COMM_H

#include "rt_ep_errorn.h"

/* Send_info is used to send data over the network.         */
/* Returns 0 if OK.                                         */
/* STA_NOT_VALID if the station is no longer on the ring.   */
/* STA_NOT_FOUND if the station isn't on the logical ring   */
/* INFO_LENGTH_OVERFLOW if the data on len is too long.     */
/* INVALID_PRIORITY if the priority is wrong (0 || >255)    */
/* INVALID_CHANNEL if the channel is not available          */
/* SYNCH_ERROR if a synchronization queue error             */
/* UNEXPECTED_ERROR if an unknown error has occurred        */

extern int send_info(char *addr_dest, unsigned short int channel_id, 
		     char *info, unsigned short int len, 
		     unsigned char priority);

/* recv_info reads from the Rx priority queue the info      */
/* to Rx, writes in *recv_info the message. The caller      */
/* process MUST assign its ID to the channel_id field       */
/* Returns:                                                 */
/* The number of Rx bytes.                                  */
/* INVALID_CHANNEL if the channel is not available          */
/* SYNCH_ERROR if a synchronization queue error             */
/* UNEXPECTED_ERROR if an unknown error has occurred        */

extern int recv_info(char *addr_sourc, unsigned short int channel_id, 
		     char *recv_info, unsigned char *priority);

/* any_info checks if there is any element in the Rx        */
/* priority queue. The caller process MUST assign its ID to */
/* the channel_id variable. Returns:                        */
/* 0 if there is.                                           */
/* NO_ELEMENTS if threre isn't                              */
/* INVALID_CHANNEL if the channel is not available          */
/* SYNCH_ERROR if a synchronization queue error             */
/* UNEXPECTED_ERROR if an unknown error has occurred        */

extern int any_info(unsigned int channel_id);

/* init_comm must be used to start the communications       */
/* thread and the tx and rx queues.                         */
/* Returns 0 if OK.                                         */
/* CREATION_ERROR if not being able creating the queues     */
/* UNEXPECTED_ERROR if an unknown error has occurred        */ 
extern int init_comm(int prio_ceiling);


#endif /*FILE_RT_COMM_H*/
